MODULE AUXILIARY

!!!!!!!!!!!!!!!!!!!!!!!!
!!   USE STATEMENTS   !!
!!!!!!!!!!!!!!!!!!!!!!!! 

USE PARAMETERS
USE GLOBAL
USE TOOLKIT
USE IFPORT

INCLUDE 'link_fnl_shared.h'

IMPLICIT NONE 

CONTAINS

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 VALUE FUNCTION ITERATION     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE iterateVF
	!! Value function iteration
	!! This routine computes the optimal policy function
	!! and the corresponding value function

	! LOCAL VARIABLES USED WITHIN LOOPS
	INTEGER                                 :: idx_prior
	INTEGER                                 :: idx_rsrv
	INTEGER                                 :: idx_sav
	INTEGER                                 :: idx_cap
	INTEGER                                 :: idx_phi
    INTEGER                                 :: idx_rhop
    INTEGER                                 :: idx_Roldmin0
    INTEGER                                 :: idx_Roldmin1
	
	INTEGER                                 :: feasible         ! contract feasibility
	LOGICAL                                 :: nocrisis         ! non crisis region
	REAL(DP)                                :: prior		    ! prior
	REAL(DP)                                :: posterior	    ! prior
    REAL(DP)                                :: res_pre		    ! prior reserves
	REAL(DP)                                :: capital		    ! capital
	REAL(DP)                                :: shock		    ! shock
	REAL(DP)                                :: res_ini		    ! initial reserves allocation
	REAL(DP)                                :: yield            ! normal gross interest rate
	REAL(DP)                                :: yield_crisis     ! crisis gross interest rate
	REAL(DP)                                :: phi_lbound       ! lower bound - normal times
	REAL(DP)                                :: phi_ubound       ! upper bound - normal times
	REAL(DP)                                :: net_y_ss         ! net output - crisis times
    REAL(DP)                                :: net_y_noss       ! net output - normal times   
    REAL(DP)                                :: net_y_autk       ! net output - autarky i.e. no feasible contract
	REAL(DP)                                :: rsav_aux         ! savings           
    REAL(DP)                                :: cons_ss          ! consumption - crisis times       
	REAL(DP)                                :: cons_noss        ! consumption - normal times
    REAL(DP)                                :: cons_autk        ! consumption - autarky i.e. no feasible contract
	REAL(DP)                                :: res_fin          ! saved reserves
    REAL(DP)                                :: Roldmin          ! minimum feasible Rold
    REAL(DP)                                :: Roldmin0         ! minimum feasible Rold (rho=0)
    REAL(DP)                                :: Roldmin1         ! minimum feasible Rold (rho=1)
	REAL(DP), DIMENSION(N_RHOP)             :: flexrhop         ! flexible RHOP grid
	REAL(DP), DIMENSION(N_RP)               :: flexrp_ss        ! flexible RP grid
    REAL(DP), DIMENSION(N_RP)               :: flexrp_noss      ! flexible RP grid
    REAL(DP), DIMENSION(N_RP)               :: flexrp_autk      ! flexible RP grid
	
	! LOCAL VARIABLES USED ACROSS/WITHIN LOOPS
	REAL(DP), DIMENSION (N_R,N_RHO)         :: valfcn_old		! last value function (W)
	REAL(DP), DIMENSION (N_R,N_RHO)         :: valfcn_new		! last value function (W_new)
	REAL(DP), DIMENSION (N_R,N_RHO)         :: policy_k		    ! auxiliary policy function (K)
	REAL(DP), DIMENSION (N_R,N_RHO)         :: policy_k_old     ! auxiliary policy function (K)
    REAL(DP), DIMENSION (N_R,N_RHO,N_PHI,N_RHOP)    :: RNEXT_OPT_old
	REAL(DP), DIMENSION (N_RP)              :: val_rp_noss      ! auxiliary value function (V_cand^N) 
    REAL(DP), DIMENSION (N_RP)              :: val_rp_ss        ! auxiliary value function (V_cand^S) 
    REAL(DP), DIMENSION (N_RP)              :: val_rp_autk      ! auxiliary value function (autarky i.e. no feasible contract) 
	REAL(DP), DIMENSION (N_PHI,N_RHOP)      :: val_aux		    ! auxiliary value - state contingent - (V)
    REAL(DP), DIMENSION (N_PHI,N_RHOP)      :: val_aux_ss	    ! auxiliary value - state contingent - (V^S)
    REAL(DP), DIMENSION (N_PHI,N_RHOP)      :: val_aux_noss	    ! auxiliary value - state contingent - (V^N)
    REAL(DP), DIMENSION (N_PHI,N_RHOP)      :: val_aux_autk     ! auxiliary value - state contingent - (autarky)
    REAL(DP), DIMENSION (N_PHI)             :: val_aux_phi      ! conditional expectation: \int (V) d \rhop
    REAL(DP), DIMENSION (N_K)               :: val_aux_k		! auxiliary value function (J): \iint (V) drhop dphi
    REAL(DP), DIMENSION (N_PHI,N_RHOP,N_K)  :: savings_aux      ! auxiliary savings decision
    REAL(DP), DIMENSION (N_PHI,N_RHOP,N_K)  :: savings_ss       ! auxiliary savings decision - crisis
    REAL(DP), DIMENSION (N_PHI,N_RHOP,N_K)  :: savings_noss     ! auxiliary savings decision - normal times
    REAL(DP), DIMENSION (N_PHI,N_RHOP)      :: savings_autk     ! auxiliary savings decision - autarky
    REAL(DP), DIMENSION (N_PHI,N_RHOP,N_K)  :: SS_aux           ! auxiliary SS decision

	! CONVERGENCE TRACKING
	INTEGER                                 :: vfi_count        ! iteration counter
	REAL(DP)                                :: conv_dist_max    ! convergence metric
	REAL(DP)                                :: conv_dist_avg    ! convergence metric
	REAL(DP)                                :: conv_dist_min    ! convergence metric
    REAL(DP)                                :: pol_dist_max     ! convergence metric
	REAL(DP)                                :: pol_dist_min     ! convergence metric
	REAL(DP)                                :: pol_dist_avg     ! convergence metric
    REAL(DP)                                :: vf_time_start
    REAL(DP)                                :: vf_time_stop

	REAL(DP)                                :: val_max          ! value function levels
	REAL(DP)                                :: val_avg          ! value function levels
	REAL(DP)                                :: val_min          ! value function levels
	
	
	! INITIALIZE
	flexrhop            = rhop_grid
	valfcn_old		    = zerodp
	vfi_count           = 1
	policy_k_old        = kmin
    phi_s_opt2          = zerodp
    rnext_opt_old       = zerodp
	
	! ITERATION
	PRINT *, ' Value function iteration ...'
	VFI_LOOP: DO

    
        vf_time_start    = DCLOCK() 
	    
		!10  FORMAT (/,1X, ' VFI iteration # ', I3, ' ... ' ) 
		! PRINT
		!WRITE (*,10)    vfi_count
        
		flexrhop    = rhop_grid
        ss_opt      = onedp
        policy_rp   = zerodp
        rnext_opt   = zerodp
        
		! outer loops - belief & incoming reserves
        !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(Roldmin, prior, res_pre, feasible, capital, res_ini, yield, yield_crisis, phi_lbound, phi_ubound, shock, nocrisis, posterior, res_fin, net_y_autk, net_y_ss, net_y_noss, flexrp_autk, flexrp_ss, flexrp_noss, savings_aux, savings_autk, cons_autk, cons_ss, cons_noss, val_aux, val_aux_k, val_aux_phi, val_aux_autk, val_aux_ss, val_aux_noss, val_rp_autk, val_rp_ss, val_rp_noss)
        !$OMP DO COLLAPSE(2)
		DO_RHO: DO idx_prior = 1,N_rho
			DO_RES: DO idx_rsrv = 1,N_r
            
            Roldmin = FSBLMIN(idx_prior)
                
                IF_AUTK: IF ( .NOT. FSBLANY(idx_prior,idx_rsrv) ) THEN   ! DOES NOT EXIST A FEASIBLE CONTRACT
                    ! In the off-equilibrium event that no contract is feasible for a given level of incoming reserves, compute value of "autarky"

                    ! retrieve values of the contract considered
                    prior           = rho_grid(idx_prior)
                    res_pre         = r_grid(idx_rsrv)                    
                    
                    ! initialize state-contingent values
                    val_aux_phi     = -infty
                    val_aux_autk    = -infty   
                    savings_autk    = zerodp

                    ! compute (state contingent) values - Even in autarky, agents observe phi shock to learn about sigma
                    DO_PHI_AUTK: DO idx_phi = 1,N_phi

                        ! retrieve the value of the shock
                        shock      = PHI_GRID(idx_phi)

                        ! net output
                        net_y_autk      = A*res_pre
                        
                        ! compute optimal savings for each posterior belief
                        DO_RHOP_AUTK: DO idx_rhop = 1,N_RHOP
                            
                            ! retrieve value of posterior
                            posterior = RHOP_GRID(idx_rhop)                            

                            ! create savings grid
                            flexrp_autk     = linspace(RMIN,net_y_autk,N_RP) 
                            
                            ! initialize V_cand
                            val_rp_autk     = -INFTY
                    
                            ! loop through candidate savings and save candidate value fcns
                            DO_SAVINGS_AUTK: DO idx_sav = 1,N_rp
                                ! candidate savings
						        res_fin     = flexrp_autk(idx_sav)
						 
						        ! compute consumption
                                cons_autk   = net_y_autk - res_fin

                                ! candidate value fcn for this node and savings w. CRISIS
						        val_rp_autk(idx_sav)  = utility(cons_autk)                      &
                                                        + BETA*INTERP2(r_grid, rho_grid,        &
                                                                        valfcn_old,             &
                                                                        N_R,N_RHO,              &
                                                                        res_fin,posterior )
                                        
                            END DO DO_SAVINGS_AUTK

                            ! update the (crisis) value/policy functions
                            val_aux_autk(idx_phi,idx_rhop)      = MAXVAL(val_rp_autk)
                            savings_autk(idx_phi,idx_rhop)      = flexrp_autk(MAXVAL(MAXLOC(val_rp_autk)))
                            
                        END DO DO_RHOP_AUTK
                            
                        ! for each local shock, compute integral across posteriors
                        val_aux_phi(idx_phi) = TRAPZ(RHOP_GRID,                         &
											    val_aux_autk(idx_phi,:)                 &
												    * RHOP_PMF(:,idx_phi,idx_prior) ,   &
												N_RHOP)
															 
                    END DO DO_PHI_AUTK           
			
				    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				    !!!! NEW VALUE FUNCTION W(reserves,prior) !!!!
				    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			
				    ! max value across feasible capital allocations
                    ! compute integral across local interim shocks
					valfcn_new(idx_rsrv,idx_prior)  = TRAPZ(PHI_GRID,                           &
												        val_aux_phi * PDF_H(PHI_GRID,prior) ,   &
												        N_PHI)
						 
				    ! compute the policy functions
                    CAP_IDX_OPT(idx_rsrv,idx_prior)     = MINVAL(MINLOC(ABS(k_grid-res_pre)))
				    policy_k(idx_rsrv,idx_prior)        = res_pre
                    PHI_S_OPT(idx_rsrv,idx_prior)       = MAXVAL(PHI_GRID)
                    PHI_N_OPT(idx_rsrv,idx_prior)       = MINVAL(PHI_GRID)
                    RL_OPT(idx_rsrv,idx_prior)          = onedp
                    RH_OPT(idx_rsrv,idx_prior)          = onedp + R
			        RNEXT_OPT(idx_rsrv,idx_prior,:,:)   = savings_autk
                
                ELSE    ! EXISTS A FEASIBLE CONTRACT
			
				    ! given prior and incoming reserves,
				    ! initialize the value function for each K
				    val_aux_k           = zerodp
			
				    ! inner loop
				    DO_KAP: DO idx_cap = 1,N_K
				
					    ! skip if feasible
					    feasible            = FSBLMAP(idx_prior,idx_rsrv,idx_cap)
					    FSBL: IF ( feasible .EQ. 1 ) THEN
					
					        ! retrieve values of the contract considered
					        prior           = rho_grid(idx_prior)
					        res_pre         = r_grid(idx_rsrv)
					        capital         = k_grid(idx_cap)
					
					        ! recover feasible contract
						    res_ini	        = FSBLRSRV(idx_prior,idx_rsrv,idx_cap)
						    yield           = FSBLR_H(idx_prior,idx_rsrv,idx_cap)
						    yield_crisis    = FSBLR_L(idx_prior,idx_rsrv,idx_cap)
						    phi_lbound      = FSBLPHI_LB(idx_prior,idx_rsrv,idx_cap)
						    phi_ubound      = FSBLPHI_UB(idx_prior,idx_rsrv,idx_cap)
						    net_y_ss        = FSBLY_SS(idx_prior,idx_rsrv,idx_cap)
                        
                            ! create savings grid (crisis)
						    flexrp_ss       = linspace(Rmin,net_y_ss,N_RP)       
                        
                            ! initialize state-contingent values
                            val_aux_phi     = -INFTY
                            val_aux         = -INFTY
                            val_aux_ss      = -INFTY
                            val_aux_noss    = -INFTY
					 
						    ! compute (state contingent) values
						    DO_PHI: DO idx_phi = 1,N_PHI
						 
						        ! retrieve the value of the shock
						        shock      = PHI_GRID(idx_phi)
                            
                                ! sudden stop region
						        nocrisis   = ( (phi_lbound .LE. shock) .AND. (shock .LE. phi_ubound) )
                            
                                ! compute the upper bound on reserves (no crisis)
							    net_y_noss  = net_output(res_pre, capital, yield, shock)
						 
							    ! create savings grid (no crisis)
							    flexrp_noss      = linspace(Rmin,net_y_noss,N_RP)   
                            
                                DO_RHOP: DO idx_rhop = 1,N_RHOP
                                    posterior = RHOP_GRID(idx_rhop)
                        
                                    !!!!!!!!!!!!!!!!!!!!!
						            !!!! sudden stop !!!!
						            !!!!!!!!!!!!!!!!!!!!!
						 
							        crisis_if: IF (.NOT. nocrisis) THEN
                                
                                        ! initialize V_cand
                                        val_rp_ss   = -INFTY
                                    
                                        ! loop through candidate savings and save candidate value fcns
                                        DO_SAVINGS_SS: DO idx_sav = 1,N_RP
								            res_fin = flexrp_ss(idx_sav)
						 
								            ! compute consumption
                                            cons_ss = net_y_ss - res_fin

                                            ! candidate value fcn for this node and savings w. CRISIS
								            val_rp_ss(idx_sav)  = utility(cons_ss)                      &
                                                                + BETA*INTERP2(r_grid, rho_grid,        &
                                                                                valfcn_old,             &
                                                                                N_R,N_RHO,              &
                                                                                res_fin,posterior )
                                        
                                        END DO DO_SAVINGS_SS
                                    
                                        ! update the (crisis) value/policy functions
                                        val_aux(idx_phi,idx_rhop)               = MAXVAL(val_rp_ss)
                                        savings_aux(idx_phi,idx_rhop,idx_cap)   = flexrp_ss(MAXVAL(MAXLOC(val_rp_ss)))

                                    END IF crisis_if
                                
                                    !!!!!!!!!!!!!!!!!!!!!!
						            !!!! NORMAL TIMES !!!!
						            !!!!!!!!!!!!!!!!!!!!!!
  
							        nocrisis_if : IF ( nocrisis ) THEN
                                    
                                        ! initialize V_cand
                                        val_rp_noss     = -INFTY
                                
                                        ! loop through candidate savings and save candidate value fcns
                                        DO_SAVINGS: DO idx_sav = 1,N_RP
								            res_fin         = flexrp_noss(idx_sav)
						 
								            ! compute consumption
								            cons_noss       = net_y_noss - res_fin
                                        
                                            val_rp_noss(idx_sav)    = utility(cons_noss)                    &
                                                                    + BETA*INTERP2(r_grid, rho_grid,        &
                                                                                    valfcn_old,             &
                                                                                    N_r,N_rho,              &
                                                                                    res_fin,posterior )
                                        END DO DO_SAVINGS
                                    
                                        ! update the (normal) value/policy functions
                                        val_aux(idx_phi,idx_rhop)               = MAXVAL(val_rp_noss)
                                        savings_aux(idx_phi,idx_rhop,idx_cap)   = flexrp_noss(MAXVAL(MAXLOC(val_rp_noss)))
                                    
                                    END IF nocrisis_if
                            
                                END DO DO_RHOP
                            
                                ! for each local shock, compute integral across posteriors
                                val_aux_phi(idx_phi) = TRAPZ(RHOP_GRID,                         &
											            val_aux(idx_phi,:)                      &
												            * RHOP_PMF(:,idx_phi,idx_prior) ,   &
												        N_RHOP)
															 
                            END DO DO_PHI               
					 
					        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
					        !!!! EXPECTED VALUE J(K) !!!!
					        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                    
					        ! compute integral across local interim shocks
					        val_aux_k(idx_cap)  = TRAPZ(PHI_GRID,                               &
												        val_aux_phi * PDF_H(PHI_GRID,prior) ,   &
												        N_PHI)
						 
					    END IF FSBL
								
				    END DO DO_KAP
			
				    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				    !!!! NEW VALUE FUNCTION W(reserves,prior) !!!!
				    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			
				    ! max value across feasible capital allocations
				    valfcn_new(idx_rsrv,idx_prior)  = MAXVAL(val_aux_k)
						 
				    ! compute the policy functions
                    cap_idx_opt(idx_rsrv,idx_prior)     = MAXVAL(MAXLOC(val_aux_k))
				    policy_k(idx_rsrv,idx_prior)        = k_grid(CAP_IDX_OPT(idx_rsrv,idx_prior))
                    phi_s_opt(idx_rsrv,idx_prior)       = FSBLPHI_UB(idx_prior,idx_rsrv,CAP_IDX_OPT(idx_rsrv,idx_prior))
                    phi_n_opt(idx_rsrv,idx_prior)       = FSBLPHI_LB(idx_prior,idx_rsrv,CAP_IDX_OPT(idx_rsrv,idx_prior))
                    rl_opt(idx_rsrv,idx_prior)          = FSBLR_L(idx_prior,idx_rsrv,CAP_IDX_OPT(idx_rsrv,idx_prior))
                    rh_opt(idx_rsrv,idx_prior)          = FSBLR_H(idx_prior,idx_rsrv,CAP_IDX_OPT(idx_rsrv,idx_prior))
			        rnext_opt(idx_rsrv,idx_prior,:,:)   = savings_aux(:,:,CAP_IDX_OPT(idx_rsrv,idx_prior))
                
                END IF IF_AUTK
                
			END DO DO_RES
        END DO DO_RHO
        !$OMP END DO
        !$OMP END PARALLEL

		!!!!!!!!!!!!!!!!!!!!!!!!!!!
		!!!! CHECK CONVERGENCE !!!!
		!!!!!!!!!!!!!!!!!!!!!!!!!!!
	  
		! compute convergence
		conv_dist_max       = MAXVAL(ABS(valfcn_old - valfcn_new) )
		conv_dist_min       = MINVAL(ABS(valfcn_old - valfcn_new) )
		conv_dist_avg       = SUM (ABS(valfcn_old - valfcn_new) ) / (N_RHO*N_R)
        
        ! compute convergence
		pol_dist_max        = MAXVAL(ABS(policy_k_old - policy_k) ) / 2                     &
                            + MAXVAL(ABS(RNEXT_OPT_old - RNEXT_OPT) ) / 2
		pol_dist_min        = MINVAL(ABS(policy_k_old - policy_k) ) / 2                     &
                            + MINVAL(ABS(RNEXT_OPT_old - RNEXT_OPT) ) / 2
		pol_dist_avg        = SUM (ABS(policy_k_old - policy_k) ) / (N_RHO*N_R) / 2         &
                            + SUM (ABS(RNEXT_OPT_old - RNEXT_OPT) ) / (N_RHO*N_R*N_PHI*N_RHOP) / 2
        
        vf_time_stop     = DCLOCK()
22      FORMAT (1X, ' took ', F15.1, ' seconds ' )  

        ! PRINT
20      FORMAT (1X, ' VF conv.      | max :   ', G10.5,' | avg :   ', G10.5)
21      FORMAT (1X, ' policy conv.  | max :   ', G10.5,' | avg :   ', G10.5)
31      FORMAT (1X, ' R1(R0=MinR0F) | (rho=1):', F8.5,' | (rho=0):', F8.5)
32      FORMAT (1X, ' PhiS(R0=MinR0F) | (rho=1):', F7.4,' | (rho=0):', F7.4)
34      FORMAT (1X, ' PhiN(R0=MinR0F) | (rho=1):', F7.4,' | (rho=0):', F7.4)
36      FORMAT (1X, ' Min R_0 FSBL  | (rho=1):', F8.5,' | (rho=0):', F8.5)
37      FORMAT (1X, ' R0next (SS)   | (rho=1):', F8.5,' | (rho=0):', F8.5)
        
        IF (trace_code) THEN
            WRITE (*,22) vf_time_stop - vf_time_start
            WRITE (*,20) conv_dist_max, conv_dist_avg
            WRITE (*,21) pol_dist_max, pol_dist_avg
            Roldmin0     = FSBLMIN(1)
            Roldmin1    = FSBLMIN(N_RHO)
            idx_Roldmin0 = MINVAL(MAXLOC(FSBLANY(1,:)))
            idx_Roldmin1 = MINVAL(MAXLOC(FSBLANY(N_RHO,:)))
            WRITE (*,31) ONEDP + Roldmin1 - policy_k(idx_Roldmin1,N_RHO), ONEDP + Roldmin0 - policy_k(idx_Roldmin0,1)
            WRITE (*,32) PHI_S_OPT(idx_Roldmin1,N_RHO), PHI_S_OPT(idx_Roldmin0,1)
            WRITE (*,34) PHI_N_OPT(idx_Roldmin1,N_RHO), PHI_N_OPT(idx_Roldmin0,1)
            WRITE (*,36) Roldmin1, Roldmin0
            WRITE (*,37) RNEXT_OPT(idx_Roldmin1,N_RHO,N_PHI,N_RHO), RNEXT_OPT(idx_Roldmin0,1,N_PHI,1)
        END IF 

777     FORMAT (1X, ' W(', I3 , ',' ,I3,') = ', G15.8)        
		
		IF (TRACE_TRACK) THEN
			DO idx_rsrv = 1,N_R
				DO idx_prior = 1,N_RHO
					WRITE (*,777) idx_rsrv,idx_prior,valfcn_new(idx_rsrv,idx_prior) 
				END DO
			END DO
		END IF 

		IF ( (conv_dist_avg .LT. CONVCRIT) .AND. (pol_dist_avg .LT. POLCRIT) ) THEN
            30  FORMAT (1X, ' CONVERGED ')
            
            ! Store policy functions
            CAP_OPT = policy_k
			
            ! Save policy functions
            OPEN (unit=9, file="CAP_OPT.out", position="rewind")
            WRITE (9, *) CAP_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="RNEXT_OPT.out", position="rewind")
            WRITE (9, *) RNEXT_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="PHI_S_OPT.out", position="rewind")
            WRITE (9, *) PHI_S_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="PHI_N_OPT.out", position="rewind")
            WRITE (9, *) PHI_N_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="RL_OPT.out", position="rewind")
            WRITE (9, *) RL_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="RH_OPT.out", position="rewind")
            WRITE (9, *) RH_OPT
            CLOSE (unit=9)
            OPEN (unit=9, file="SS_OPT.out", position="rewind")
            WRITE (9, *) SS_opt
            CLOSE (unit=9)
                
			EXIT
        ELSE IF (vfi_count .GT. VF_IT_MAX) THEN
            40  FORMAT (1X, ' OUT OF ITERATIONS ',/) 
			! PRINT
			WRITE (*,40)             
            EXIT
		END IF
	
		! increment counter
		vfi_count       = vfi_count + 1 
		
		! compute updated value function guess
		valfcn_old      = WEIGHT*valfcn_old + (1-WEIGHT)*valfcn_new
	
		! update last policy function
		policy_k_old    = policy_k
        RNEXT_OPT_old    = RNEXT_OPT
	
    END DO VFI_LOOP
    
    
				
END SUBROUTINE iterateVF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 POSTERIOR PMF LOOKUP     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE createPMF
	!! This creates the PMF of posterior
    !! and stores them in a table
    !! This speeds up the code by computing
    !! the ALTG once and for all.
    
	! LOCAL VARIABLES USED WITHIN LOOPS
	INTEGER                                 :: idx_prior
	INTEGER                                 :: idx_rhop
	INTEGER                                 :: idx_phi
	
    ! PROGRAM
    FORALL (idx_rhop=1:N_RHOP, idx_phi=1:N_PHI, idx_prior=1:N_RHO)
        RHOP_PMF(idx_rhop,idx_phi,idx_prior) = ALTG(RHOP_GRID(idx_rhop),    &
                                                    PHI_GRID(idx_phi),      &
                                                    rho_grid(idx_prior))
    END FORALL
   
    DO_RHO: DO idx_prior = 1,N_RHO
        DO_PHI: DO idx_phi = 1,N_PHI
            RHOP_PMF_SUM(idx_phi,idx_prior) = TRAPZ(                            &
                                                RHOP_GRID,                      &
                                                RHOP_PMF(:,idx_phi,idx_prior) , &
                                                N_RHOP)
        END DO DO_PHI

        RHOP_PMF_MIN    = MINVAL (RHOP_PMF_SUM, DIM = 1)
        RHOP_PMF_MAX    = MAXVAL (RHOP_PMF_SUM, DIM = 1)
        RHOP_PMF_AVG    = SUM (RHOP_PMF_SUM, DIM = 1)/N_PHI

10      FORMAT (1X, 'ALTG check pre  :  rho = ', F7.3, '; max = ', F7.3,' ; ',' min = ', F7.3,' avg = ', F7.3)
        ! PRINT
        IF (TRACE_TRACK) THEN
            WRITE (*,10) rho_grid(idx_prior), RHOP_PMF_MAX(idx_prior), RHOP_PMF_MIN(idx_prior), RHOP_PMF_AVG(idx_prior)
        END IF
        
    END DO DO_RHO

    
    ! Fixing the CDF issue manually
    IF_FORCE: IF (CDF_FORCE) THEN
        
        FORALL (idx_phi=1:N_PHI, idx_prior=1:N_RHO)
            ! renormalize CDF
            RHOP_PMF(:,idx_phi,idx_prior)  = RHOP_PMF(:,idx_phi,idx_prior)/ &
                                                RHOP_PMF_SUM(idx_phi,idx_prior)
        END FORALL    

        DO_RHO2: DO idx_prior = 1,N_RHO
            DO_PHI2: DO idx_phi = 1,N_PHI
                RHOP_PMF_SUM(idx_phi,idx_prior) = TRAPZ(                            &
                                                    RHOP_GRID,                      &
                                                    RHOP_PMF(:,idx_phi,idx_prior) , &
                                                    N_RHOP)
            END DO DO_PHI2

            RHOP_PMF_MIN    = MINVAL (RHOP_PMF_SUM, DIM = 1)
            RHOP_PMF_MAX    = MAXVAL (RHOP_PMF_SUM, DIM = 1)
            RHOP_PMF_AVG    = SUM (RHOP_PMF_SUM, DIM = 1)/N_PHI

11          FORMAT (1X, 'ALTG check post :  rho = ', F7.3, '; max = ', F7.3,' ; ',' min = ', F7.3,' avg = ', F7.3)
            ! PRINT
            IF (TRACE_TRACK) THEN
                WRITE (*,11) rho_grid(idx_prior), RHOP_PMF_MAX(idx_prior), RHOP_PMF_MIN(idx_prior), RHOP_PMF_AVG(idx_prior)
            END IF

        END DO DO_RHO2   

    END IF IF_FORCE
				
END SUBROUTINE createPMF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	    MAKE REGIONS          !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE makeRegions

    ! LOCAL VARIABLES
	INTEGER     :: idx_eme

    INTEGER     :: idx_reg          ! index for region
    INTEGER     :: idx_pos          ! index for country position within region

    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! CONSTRUCT REGIONS !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    
100  FORMAT (1X,                    &
            "eme # ",       T10,    &
            "reg # ",       T20,    &
            "pos # ",       T30     )
     
102  FORMAT (1X,                    &
            I3,             T10,    &
            I3,             T20,    &
            I3,             T30     )
     
    IF (TRACE_TRACK) THEN
        WRITE(*,100)
    END IF
    
    DO idx_eme = 1,N_ALL_EME
        
        ! get the position within a region = 1 ... N_EME
        idx_pos     = MODULO(idx_eme - 1, N_EME) + 1
        ! get the region index = 1 ... N_REG
        idx_reg     = FLOOR( REAL(idx_eme - 1)/REAL(N_EME) ) + 1
        
        ! assign country to region
        REGIONS(idx_eme) =  idx_reg
        PEERS(idx_reg, idx_pos) =  idx_eme
        
        IF (TRACE_TRACK) THEN
            WRITE(*,102) idx_eme, idx_reg, idx_pos
        END IF
        
    END DO

END SUBROUTINE makeRegions

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!          SIMULATION            !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE simulation (simul_results)
	!! Simulates the model
    
	! INPUT & OUTPUT
	TYPE(RESULTS_SIM)           simul_results    
    INTENT(INOUT)           ::  simul_results    
    OPTIONAL                ::  simul_results    

    
	! LOCAL VARIABLES USED WITHIN LOOPS
	INTEGER     :: idx_simul
	INTEGER     :: idx_time
    INTEGER     :: idx_eme
    INTEGER     :: idx_phi
    INTEGER     :: idx_Roldmin
    INTEGER     :: idx_Rold
    INTEGER     :: idx_rho
	
    REAL(DP)    :: Rold
    REAL(DP)    :: prior
    REAL(DP)    :: shock
    
    REAL(DP)    :: R_pre
    REAL(DP)    :: R_post
    REAL(DP)    :: R_trans
    REAL(DP)    :: SS_count_pre
    REAL(DP)    :: SS_count_post
    REAL(DP)    :: SS_count_trans
    REAL(DP)    :: SS_count_cut
    REAL(DP)    :: haircut
    REAL(DP)    :: haircut_cond
    REAL(DP)    :: SS_prob_pre
    REAL(DP)    :: SS_prob_trans
    REAL(DP)    :: SS_prob_post
    REAL(DP)    :: SS_prob_nontrans
    REAL(DP)    :: Rold_avg
    REAL(DP)    :: Rold_pre
    REAL(DP)    :: Rold_post
    REAL(DP)    :: Rold_trans
    REAL(DP)    :: phiS_pre
    REAL(DP)    :: phiS_trans
    REAL(DP)    :: phiS_post
    REAL(DP)    :: phiN_pre
    REAL(DP)    :: phiN_trans
    REAL(DP)    :: phiN_post
    REAL(DP)    :: RH_pre
    REAL(DP)    :: RH_trans
    REAL(DP)    :: RH_post
    REAL(DP)    :: beliefs_pre          
    REAL(DP)    :: beliefs_trans     
    REAL(DP)    :: beliefs_post 
    
    REAL(DP), DIMENSION (T_END)     :: R1_ts
    REAL(DP), DIMENSION (T_END)     :: R0_ts
    REAL(DP), DIMENSION (T_END)     :: SS_count_ts
    
    REAL(DP)    :: SS_simul_pre     ! counts SS in a simulation x pre era
    REAL(DP)    :: SS_simul_trans   ! counts SS in a simulation x trans era
    REAL(DP)    :: SS_simul_post    ! counts SS in a simulation x post era
    INTEGER     :: idx_hist
    REAL(DP), DIMENSION (N_HIST)    :: SS_hist_pre
    REAL(DP), DIMENSION (N_HIST)    :: SS_hist_trans
    REAL(DP), DIMENSION (N_HIST)    :: SS_hist_post
    
    INTEGER     :: idx_reg          ! index for region
    INTEGER     :: idx_pos          ! index for country position within region
    
    ! PROGRAM
    PRINT *, ' Simulating ...'
    ! Read policy functions
    OPEN (unit=9, file="CAP_OPT.out", position="rewind")
    READ (9, *) CAP_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="RNEXT_OPT.out", position="rewind")
    READ (9, *) RNEXT_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="PHI_S_OPT.out", position="rewind")
    READ (9, *) PHI_S_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="PHI_N_OPT.out", position="rewind")
    READ (9, *) PHI_N_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="RL_OPT.out", position="rewind")
    READ (9, *) RL_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="RH_OPT.out", position="rewind")
    READ (9, *) RH_OPT
    CLOSE (unit=9)
    OPEN (unit=9, file="SS_OPT.out", position="rewind")
    READ (9, *) SS_opt
    CLOSE (unit=9)
    
    !!!!!!!!!!!!!!!!!!!!!!!!
    ! initializing objects !
    !!!!!!!!!!!!!!!!!!!!!!!!   
    BELIEFS_SIMUL   = RHO_INIT
    PHI_SIMUL       = ZERODP
    idx_Roldmin     = MINVAL(MAXLOC(FSBLANY(N_rho,:)))
    ROLD_SIMUL      = R_GRID(idx_Roldmin)
    K_SIMUL         = ZERODP
    SS_SIMUL        = ZERODP
    HAIRCUT_SIMUL   = ZERODP
    PHIR_SIMUL      = ZERODP
    PHIS_SIMUL      = ZERODP
    PHIN_SIMUL      = ZERODP
    RH_SIMUL        = ZERODP
    RL_SIMUL        = ZERODP
    RPRIME_SIMUL    = ZERODP
    C_SIMUL         = ZERODP
    
    !! Generate and store array of random numbers
    CALL DRNUN(N_S*N_ALL_EME*T_END,PHI_SIMUL)
        
    !! Transform random numbers to phi shocks from Bounded Pareto Distribution
    ! sigma_L if t<T_change
    PHI_SIMUL(:,:,1:T_CHANGE-1)     = - (PHI_SIMUL(:,:,1:T_CHANGE-1)**SIGMA_L - 1) 
    ! sigma_H if t>=T_change
    PHI_SIMUL(:,:,T_CHANGE:T_END)   = - (PHI_SIMUL(:,:,T_CHANGE:T_END)**SIGMA_H - 1) 
        
8   FORMAT (1X,/' sigma_L: ', F8.5, ' | Avg phi shock: ', F8.5,' | Expected: ', F8.5)
9   FORMAT (1X,  'sigma_H: ', F8.5, ' | Avg phi shock: ', F8.5,' | Expected: ', F8.5/)
    
    IF (trace_code) THEN
        WRITE(*,8) SIGMA_L,                                             &
                    SUM(SUM(SUM(PHI_SIMUL(:,:,1:T_CHANGE-1),3),2),1)/   &
                        (N_S*N_EME*T_CHANGE),                           &
                    SIGMA_L/(1+SIGMA_L)
    
        WRITE(*,9) SIGMA_H, &
                    SUM(SUM(SUM(PHI_SIMUL(:,:,T_CHANGE:T_END),3),2),1)/ &
                        (N_S*N_EME*(T_END-T_CHANGE+1)),                 &
                    SIGMA_H/(1+SIGMA_H)
    END IF
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! CONSTRUCT REGIONS !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    CALL makeRegions
        
    
    !!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! RUN SIMULATIONS !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!
    
    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(prior, idx_reg, shock, Rold, idx_phi)
    !$OMP DO COLLAPSE(3)
    DO_SIMUL: DO idx_simul = 1,N_S        
        DO_TIME: DO idx_time = 1,T_END            
            DO_EME: DO idx_eme = 1,N_ALL_EME
                
                ! retrieve prior
                prior   = BELIEFS_SIMUL(idx_simul,idx_eme,idx_time)
                
                ! get region
                idx_reg = REGIONS(idx_eme)
            
                ! Update beliefs according to Bayes' Rule
                IF (idx_time < T_end) THEN
                    
                    BELIEFS_SIMUL(idx_simul,idx_eme,idx_time+1) =                           &
                        MAX( RHOMIN, MIN( RHOMAX,                                           &
                            1 / ( 1 + (1-prior)/prior                                       &
                                    * PRODVEC( pdf_sigma_H( phi_simul(idx_simul,            &
                                                                PEERS(idx_reg,:),         &   ! peer group only
                                                                idx_time)                   &
                                                                ), N_EME )                  &
                                    / PRODVEC( pdf_sigma_L( phi_simul(idx_simul,            &
                                                                PEERS(idx_reg,:),         &   ! peer group only
                                                                idx_time)                   &
                                                                ), N_EME )                  &
                                )                                                           &
                        ) )
                    
                END IF                
                
                shock   = PHI_SIMUL(idx_simul,idx_eme,idx_time)
                Rold    = ROLD_SIMUL(idx_simul,idx_eme,idx_time)
                idx_Rold = MAXVAL(MINLOC(ABS(R_grid-Rold)))
                idx_rho = MAXVAL(MINLOC(ABS(rho_grid-prior)))
                
                IF (FSBLANY(idx_rho,idx_Rold)) THEN
                    K_SIMUL(idx_simul,idx_eme,idx_time) = &
                        interp2(r_grid,rho_grid,CAP_OPT,N_R,N_RHO,Rold,prior)
                    PHIR_SIMUL(idx_simul,idx_eme,idx_time) = Rold + onedp - K_SIMUL(idx_simul,idx_eme,idx_time)
                    PHIS_SIMUL(idx_simul,idx_eme,idx_time) = &
                        interp2(r_grid,rho_grid,PHI_S_OPT,N_R,N_RHO,Rold,prior)
                    PHIN_SIMUL(idx_simul,idx_eme,idx_time) = &
                        interp2(r_grid,rho_grid,PHI_N_OPT,N_R,N_RHO,Rold,prior)
                    RH_SIMUL(idx_simul,idx_eme,idx_time) = &
                        interp2(r_grid,rho_grid,RH_OPT,N_R,N_RHO,Rold,prior)
                ELSE ! If no feasible contract exists, simulate autarky
                    K_SIMUL(idx_simul,idx_eme,idx_time) = Rold
                    PHIR_SIMUL(idx_simul,idx_eme,idx_time) = zerodp
                    PHIS_SIMUL(idx_simul,idx_eme,idx_time) = PHI_S_OPT(idx_Rold,idx_rho)
                    PHIN_SIMUL(idx_simul,idx_eme,idx_time) = PHI_N_OPT(idx_Rold,idx_rho)
                    RH_SIMUL(idx_simul,idx_eme,idx_time) = RH_OPT(idx_Rold,idx_rho)
                END IF
                
                ! Sudden stop if shock > phiS
                IF ((shock > PHIS_SIMUL(idx_simul,idx_eme,idx_time)).OR.(shock < PHIN_SIMUL(idx_simul,idx_eme,idx_time))) THEN
                    SS_SIMUL(idx_simul,idx_eme,idx_time) = ONEDP
                    HAIRCUT_SIMUL(idx_simul,idx_eme,idx_time) = ONEDP - &
                        interp2(r_grid,rho_grid,RL_OPT,N_R,N_RHO,Rold,prior)
                END IF
                
                ! index of phi shock
                idx_phi = MINLOC( ABS( shock - PHI_GRID ) , 1 )
                
                ! Saved Reserves
                IF (idx_time < T_end) THEN
                    ROLD_SIMUL(idx_simul,idx_eme,idx_time+1) =          &
                        interp3(r_grid,rho_grid,RHOP_GRID,              &
                        RNEXT_OPT(:,:,idx_phi,:), N_R, N_RHO, N_RHOP,   &
                        Rold, prior,                                    &
                        BELIEFS_SIMUL(idx_simul,idx_eme,idx_time+1) )
                END IF
                
            END DO DO_EME
        END DO DO_TIME
    END DO DO_SIMUL    
    !$OMP END DO
    !$OMP END PARALLEL
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! Compute averages !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!
    
    Rold_avg = SUM( SUM( SUM(                           &   
                ROLD_SIMUL(:,:,T_START:T_END)           &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_END-T_START+1)
    
    Rold_pre = SUM( SUM( SUM(                           &   
                ROLD_SIMUL(:,:,T_START:T_CHANGE-1)      &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_CHANGE-T_START)
    
    Rold_trans = SUM( SUM( SUM(                             &   
                ROLD_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1) &
                ,3) ,2) ,1)                                 &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_trans)
    
    Rold_post = SUM( SUM( SUM(                          &   
                ROLD_SIMUL(:,:,T_CHANGE+q_trans:T_END)  &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_END-T_CHANGE-q_trans+1)
    
    beliefs_pre = SUM( SUM( SUM(                           &   
                BELIEFS_SIMUL(:,:,T_START:T_CHANGE-1)      &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_CHANGE-T_START)        
    beliefs_trans = SUM( SUM( SUM(                             &   
                BELIEFS_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1) &
                ,3) ,2) ,1)                                 &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_trans)
    beliefs_post = SUM( SUM( SUM(                          &   
                BELIEFS_SIMUL(:,:,T_CHANGE+q_trans:T_CHANGE+q_trans+q_post-1)  &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_post)
    
    R_pre = SUM( SUM( SUM(                              &   
                PHIR_SIMUL(:,:,T_START:T_CHANGE-1)      &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_CHANGE-T_START)                  
    
    R_trans = SUM( SUM( SUM(                                &   
                PHIR_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1) &
                ,3) ,2) ,1)                                 &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_trans)
        
    R_post = SUM( SUM( SUM(                             &   
                PHIR_SIMUL(:,:,T_CHANGE+q_trans:T_END)  &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_END-T_CHANGE-q_trans+1)

    SS_count_pre = SUM( SUM( SUM(                       &   
                SS_SIMUL(:,:,T_CHANGE-q_pre:T_CHANGE-1) &
                ,3) ,2) ,1)/DBLE(N_S)
    
    SS_count_trans = SUM( SUM( SUM(                         &   
                SS_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1)   &
                ,3) ,2) ,1)/DBLE(N_S)
    
    SS_count_post = SUM( SUM( SUM(                                          &   
                SS_SIMUL(:,:,T_CHANGE+q_trans:T_CHANGE+q_trans+q_post-1)    &
                ,3) ,2) ,1)/DBLE(N_S)
    
    phiS_pre = SUM( SUM( SUM(                           &   
                PHIS_SIMUL(:,:,T_START:T_CHANGE-1)      &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_CHANGE-T_START)                  
    
    phiS_trans = SUM( SUM( SUM(                             &   
                PHIS_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1) &
                ,3) ,2) ,1)                                 &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_trans)
        
    phiS_post = SUM( SUM( SUM(                          &   
                PHIS_SIMUL(:,:,T_CHANGE+q_trans:T_END)  &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_END-T_CHANGE-q_trans+1)
    
    phiN_pre = SUM( SUM( SUM(                           &   
                PHIN_SIMUL(:,:,T_START:T_CHANGE-1)      &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_CHANGE-T_START)                  
    
    phiN_trans = SUM( SUM( SUM(                             &   
                PHIN_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1) &
                ,3) ,2) ,1)                                 &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(q_trans)
        
    phiN_post = SUM( SUM( SUM(                          &   
                PHIN_SIMUL(:,:,T_CHANGE+q_trans:T_END)  &
                ,3) ,2) ,1)                             &
                /DBLE(N_S)/DBLE(N_ALL_EME)/DBLE(T_END-T_CHANGE-q_trans+1)

    SS_prob_pre     = SS_count_pre / DBLE(q_pre) / DBLE(N_ALL_EME)
    SS_prob_trans   = SS_count_trans / DBLE(q_trans) / DBLE(N_ALL_EME)
    SS_prob_post    = SS_count_post / DBLE(q_post) / DBLE(N_ALL_EME)
    SS_prob_nontrans    = ( SS_count_pre + SS_count_post ) / DBLE( q_pre + q_post ) / DBLE(N_ALL_EME)
                        
    RH_pre = SUM( SUM(                                                                      &
                PRODUCT(RH_SIMUL(:,:,T_START:T_CHANGE-1),3)**(onedp/DBLE(T_CHANGE-T_START)) &
                ,2) ,1)/DBLE(N_S)/DBLE(N_ALL_EME)
    
    RH_trans = SUM( SUM(                                                                        &
                PRODUCT(RH_SIMUL(:,:,T_CHANGE:T_CHANGE+q_trans-1),3)**(onedp/DBLE(q_trans))     &
                ,2) ,1)/DBLE(N_S)/DBLE(N_ALL_EME)
    
    RH_post = SUM( SUM(                                                                                         &
                PRODUCT(RH_SIMUL(:,:,T_CHANGE+q_trans:T_END),3)**(onedp/DBLE(T_END-T_CHANGE-q_trans+1)) &
                ,2) ,1)/DBLE(N_S)/DBLE(N_ALL_EME)
    
    haircut         = ZERODP
    SS_count_cut    = ZERODP
    haircut_cond    = ZERODP
    
    DO idx_simul = 1,N_S
        DO idx_time = T_CHANGE - q_pre, T_CHANGE + q_trans + q_post - 1
            DO idx_eme = 1,N_ALL_EME
                IF ( SS_SIMUL(idx_simul,idx_eme,idx_time) > 0 ) THEN
                    haircut = haircut + HAIRCUT_SIMUL(idx_simul,idx_eme,idx_time)
                    IF (HAIRCUT_SIMUL(idx_simul,idx_eme,idx_time) > ZERODP) THEN
                        SS_count_cut = SS_count_cut + 1
                        haircut_cond = haircut_cond + HAIRCUT_SIMUL(idx_simul,idx_eme,idx_time)
                    END IF
                END IF
            END DO
        END DO
    END DO

    haircut = haircut / DBLE( SS_count_pre + SS_count_trans + SS_count_post ) / DBLE(N_S)
    haircut_cond = haircut_cond / DBLE( SS_count_cut )
    
    DO idx_time = 1,T_END
        SS_count_ts(idx_time) = SUM( SUM( SS_SIMUL(:,:,idx_time) ,2) ,1)/DBLE(N_S)
        R1_ts(idx_time) = SUM( SUM( PHIR_SIMUL(:,:,idx_time) ,2) ,1)/DBLE(N_S)/DBLE(N_ALL_EME)
        R0_ts(idx_time) = SUM( SUM( ROLD_SIMUL(:,:,idx_time) ,2) ,1)/DBLE(N_S)/DBLE(N_ALL_EME)
    END DO
      
    
    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! compute histogram !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    !! initialize histogram vector
    SS_hist_pre     = zerodp
    SS_hist_trans   = zerodp
    ss_hist_post    = zerodp
    
    DO_HIST: DO idx_simul = 1,N_S
        SS_simul_pre     = SUM( SUM(                                        &   
                        SS_SIMUL(IDX_SIMUL,:,T_CHANGE-q_pre:T_CHANGE-1)     &
                        ,2) ,1)

        SS_simul_trans   = SUM( SUM(                                        &   
                        SS_SIMUL(IDX_SIMUL,:,T_CHANGE:T_CHANGE+q_trans-1)   &
                        ,2) ,1)

        SS_simul_post   = SUM( SUM(                                                     &   
                        SS_SIMUL(IDX_SIMUL,:,T_CHANGE+q_trans:T_CHANGE+q_trans+q_post)  &
                        ,2) ,1)
        
        IF ( SS_simul_pre .LT. (N_HIST-1) ) THEN
            idx_hist    = INT(SS_simul_pre) + 1
            SS_hist_pre(idx_hist)   = SS_hist_pre(idx_hist) + onedp
        ELSE
            SS_hist_pre(N_HIST)     = SS_hist_pre(N_HIST) + onedp
        END IF
        
        IF ( SS_simul_trans .LT. (N_HIST-1) ) THEN
            idx_hist    = INT(SS_simul_trans) + 1
            SS_hist_trans(idx_hist) = SS_hist_trans(idx_hist) + onedp
        ELSE
            SS_hist_trans(N_HIST)   = SS_hist_trans(N_HIST) + onedp
        END IF

        IF ( SS_simul_post .LT. (N_HIST-1) ) THEN
            idx_hist    = INT(SS_simul_post) + 1
            SS_hist_post(idx_hist) = SS_hist_post(idx_hist) + onedp
        ELSE
            SS_hist_post(N_HIST)   = SS_hist_post(N_HIST) + onedp
        END IF
    END DO DO_HIST 
    
    !! convert frequencies into histogram
    SS_hist_pre     = SS_hist_pre/DBLE(N_S)
    SS_hist_trans   = SS_hist_trans/DBLE(N_S)
    SS_hist_post    = SS_hist_post/DBLE(N_S)

    
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !!!! Report simulation results !!!!
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
20  FORMAT (1X, 'Avg Reserves | pre: ', F8.5,' | trans: ', F8.5,' | post: ', F8.5)
    WRITE(*,20) R_pre, R_trans, R_post

21  FORMAT (1X, 'Sudden Stops | pre: ', F8.3,' | trans: ', F8.3,' | post: ', F8.3)
    WRITE(*,21) SS_count_pre, SS_count_trans, SS_count_post
    
22  FORMAT (1X, 'SS prob (%)  | pre: ', F8.3,' | trans: ', F8.3,' | post: ', F8.3)
    WRITE(*,22) SS_prob_pre*100.0_DP, SS_prob_trans*100.0_DP, SS_prob_post*100.0_DP
    
26  FORMAT (1X, '             | non-transition: ', F8.3)
    WRITE(*,26) SS_prob_nontrans*100.0_DP
    
23  FORMAT (1X, 'Average SS Haircut (%): ', F10.3)
    WRITE(*,23) haircut*100.0_DP
        
    IF (trace_code) THEN

24      FORMAT (1X, 'Rold         | pre: ', F8.3,' | trans: ', F8.3,' | post: ', F8.3)
        WRITE(*,24) Rold_pre, Rold_trans, Rold_post
    
25      FORMAT (1X, 'RH           | pre: ', F8.5,' | trans: ', F8.5,' | post: ', F8.5)
        WRITE(*,25) RH_pre, RH_trans, RH_post
                
19      FORMAT (1X, 'Avg PhiS     | pre: ', F8.5,' | trans: ', F8.5,' | post: ', F8.5)
        WRITE(*,19) phiS_pre, phiS_trans, phiS_post    
    
31      FORMAT (1X, 'Avg PhiN     | pre: ', F8.5,' | trans: ', F8.5,' | post: ', F8.5)
        WRITE(*,31) phiN_pre, phiN_trans, phiN_post
        
27      FORMAT (1X, 'Average SS Haircut (%): ', F10.3, ' (conditional on haircut) ')
        WRITE(*,27) haircut_cond*100.0_DP
    
28      FORMAT (1X, 'Haircut probability(%): ', F10.3, ' (conditional on SS) ')
        WRITE(*,28) SS_count_cut/N_S/( SS_count_pre + SS_count_trans + SS_count_post )*100.0_DP
    
40      FORMAT (1X, 'Erg. Dist. (pre) | R_0min: ', F7.5, ' | R_0max: ', F7.5/)
        WRITE(*,40) MINVAL(Rold_simul(:,:,T_START:T_CHANGE-1)), MAXVAL(Rold_simul(:,:,T_START:T_CHANGE-1))
    
41      FORMAT (1X, 'Erg. Dist. (trans) | R_0min: ', F7.5, ' | R_0max: ', F7.5/)
        WRITE(*,41) MINVAL(Rold_simul(:,:,T_CHANGE:T_CHANGE+19)), MAXVAL(Rold_simul(:,:,T_CHANGE:T_CHANGE+19))
    
42      FORMAT (1X, 'Erg. Dist. (post) | R_0min: ', F7.5, ' | R_0max: ', F7.5/)
        WRITE(*,42) MINVAL(Rold_simul(:,:,T_CHANGE+20:T_END)), MAXVAL(Rold_simul(:,:,T_CHANGE+20:T_END))
    END IF
        

29  FORMAT (1X, 'Sudden stops: ', / &
            1X,'  N',T5,'   PRE   ', T15,'  TRANS  ', T25,'   POST  ') 
    WRITE(*,29)    
30  FORMAT (1X,I3, T5, F8.5, T15, F8.5, T25, F8.5)
    
    DO idx_hist = 1,N_HIST
        WRITE(*,30) idx_hist-1,                 &
                    SS_hist_pre(idx_hist),      &
                    SS_hist_trans(idx_hist),    &
                    SS_hist_post(idx_hist)
    END DO    
    
    
    OPEN (unit=9, file="results.out", position="rewind")
    WRITE(9,20) R_pre, R_trans, R_post
    WRITE(9,19) phiS_pre, phiS_trans, phiS_post
    WRITE(9,21) SS_count_pre, SS_count_trans, SS_count_post
    WRITE(9,22) SS_prob_pre*100.0_DP, SS_prob_trans*100.0_DP, SS_prob_post*100.0_DP
    WRITE(9,26) 1-cdf_sigma_L(phiS_pre), 1-cdf_sigma_H(phiS_trans), 1-cdf_sigma_H(phiS_post)
    WRITE(9,25) RH_pre, RH_trans, RH_post
    WRITE(9,23) haircut*100.0_DP
    WRITE(9,24) Rold_avg 
    WRITE(9,29)  
    DO idx_hist = 1,N_HIST
        WRITE(9,30) idx_hist-1,     &
        SS_hist_pre(idx_hist),      &
        SS_hist_trans(idx_hist),    &
        SS_hist_post(idx_hist)
    END DO
    CLOSE (unit=9)
    
    ! Print time series of reserves and SS (averaged over country and simulations) 
    OPEN (unit=9, file="R1_ts.out", position="rewind")
    WRITE (9, *) R1_ts
    CLOSE (unit=9)
    
    OPEN (unit=9, file="R0_ts.out", position="rewind")
    WRITE (9, *) R0_ts
    CLOSE (unit=9)
    
    OPEN (unit=9, file="SS_count_ts.out", position="rewind")
    WRITE (9, *) SS_count_ts
    CLOSE (unit=9)
    
    ! Print simulation of reserves and SS
        
300 FORMAT (  1X,                   &
            'simul    ',    T10,    &
            'country  ',    T20,    &
            'time     ',    T30,    &
            'reserves ',    T40,    &
            'SS       '             &
                )

302 FORMAT (  1X,                   &
            I5,             T10,    &
            I5,             T20,    &
            I5,             T30,    &
            F9.5,           T40,    &
            F5.1                    &
                )

    OPEN (unit=9, file="SIMULATIONS.out", position="rewind")
    WRITE (9,300)
    DO idx_simul = 1,N_S
        DO idx_eme = 1,N_ALL_EME
            DO idx_time = T_ERG+1,T_END
                WRITE (9,302) idx_simul, idx_eme, idx_time-T_ERG,    &
                            PHIR_SIMUL(idx_simul,idx_eme,idx_time), &
                            SS_SIMUL(idx_simul,idx_eme,idx_time)
            END DO
        END DO
    END DO
    CLOSE (unit=9)
    
    
    
    IF (TRACE_TRACK) THEN        
        WRITE(*,*) 'Sample of Saved Reserves'
        WRITE(*,*) Rold_simul(1,1,2:T_end)
    END IF

    IF ( PRESENT(simul_results) ) THEN
        
        simul_results%BETA_VAL          = BETA
        simul_results%THETA_VAL         = THETA
        simul_results%SIGMA_L_VAL       = SIGMA_L
        simul_results%SIGMA_H_VAL       = SIGMA_H
        
        simul_results%R_INI_PRE         = R_pre
        simul_results%R_INI_TRANS       = R_trans
        simul_results%R_INI_POST        = R_post
        
        simul_results%rho_pre           = beliefs_pre
        simul_results%rho_trans         = beliefs_trans
        simul_results%rho_post          = beliefs_post

        simul_results%R_OLD_AVG         = Rold_avg
        
        simul_results%SS_FREQ_PRE       = SS_prob_pre*100.0_DP
        simul_results%SS_FREQ_TRANS     = SS_prob_trans*100.0_DP
        simul_results%SS_FREQ_POST      = SS_prob_post*100.0_DP
        simul_results%SS_FREQ_NONTRANS  = SS_prob_nontrans*100.0_DP
        
        simul_results%R_H_PRE           = RH_pre
        simul_results%R_H_TRANS         = RH_trans
        simul_results%R_H_POST          = RH_post
        
        simul_results%HAIRCUT_ALL       = haircut*100.0_DP
        simul_results%HAIRCUT_COND      = haircut_cond*100.0_DP
        
        simul_results%PHI_S_PRE         = phiS_pre
        simul_results%PHI_S_TRANS       = phiS_trans
        simul_results%PHI_S_POST        = phiS_post
        
        simul_results%SS_COUNT_PRE      = SS_count_pre
        simul_results%SS_COUNT_TRANS    = SS_count_trans
        simul_results%SS_COUNT_POST     = SS_count_post
        
        simul_results%SS_HIST_PRE       = SS_hist_pre
        simul_results%SS_HIST_TRANS     = SS_hist_trans
        simul_results%SS_HIST_POST      = SS_hist_post      
        
        simul_results%Rold_erg_min      = MINVAL(Rold_simul)
        simul_results%Rold_erg_max      = MAXVAL(Rold_simul)
        
    END IF    
    
    END SUBROUTINE simulation

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!          RUN BASELINE           !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE baseline

    ! LOCAL VARIABLES
    INTEGER                                 :: idx_beta
    INTEGER                                 :: idx_theta
    INTEGER                                 :: idx_sigma_l
    INTEGER                                 :: idx_sigma_h
    INTEGER                                 :: idx_hist
    REAL(DP)                                :: simul_num
    REAL(DP)                                :: simul_tot
    TYPE(RESULTS_SIM)                       simul_results

    ! BUILD PARAMETER GRID FOR BETA
    CALL creategrid(BETA_GRID,N_BETA,BETAMIN,BETAMIN,BETAMAX,ONEPWR)

    ! BUILD PARAMETER GRID FOR THETA
    CALL creategrid(THETA_GRID,N_THETA,THETAMIN,THETAMIN,THETAMAX,ONEPWR)

    ! BUILD PARAMETER GRID FOR SIGMA_L
    CALL creategrid(SIGMA_L_GRID,N_SIGMA_L,SIGMA_LMIN,SIGMA_LMIN,SIGMA_LMAX,ONEPWR)

    ! BUILD PARAMETER GRID FOR SIGMA_H
    CALL creategrid(SIGMA_H_GRID,N_SIGMA_H,SIGMA_HMIN,SIGMA_HMIN,SIGMA_HMAX,ONEPWR)

    
100  FORMAT (/,1X, ' SOLVING THE MODEL - START ') 
	 
	! PRINT
	WRITE (*,100)
    
    
    ! Print key moments
200 FORMAT (  1X,                   &
            'beta   ',      T10,    &
            'theta    ',    T20,    &
            'sigma_l  ',    T30,    &
            'sigma_h  ',    T40,    &
            'res pre  ',    T50,    &
            'res tra  ',    T60,    &
            'res pos  ',    T70,    &
            'res sav  ',    T80,    &
            'ss % pre ',    T90,   &
            'ss % tra ',    T100,   &
            'ss % pos ',    T110,   &
            'ss % nont',    T120,   &
            'R_H pre  ',    T130,   &
            'R_H tra  ',    T140,   &
            'R_H pos  ',    T150,   &
            'hcut all ',    T160,   &
            'hcut con ',    T170,   &
            'phiS pre ',    T180,   &
            'phiS tra ',    T190,   &
            'phiS pos ',    T200,   &
            'MIN erR_0',    T210,   &
            'MAX erR_0',    T220,   &
            'rho pre  ',    T230,   &
            'rho trans',    T240,   &
            'rho post '             &
                )
    
201             FORMAT (  1X,                   &
            '=======',      T10,    &
            '=========',    T20,    &
            '=========',    T30,    &
            '=========',    T40,    &
            '=========',    T50,    &
            '=========',    T60,    &
            '=========',    T70,    &
            '=========',    T80,    &
            '=========',    T90,    &
            '=========',    T100,   &
            '=========',    T110,   &
            '=========',    T120,   &
            '=========',    T130,   &
            '=========',    T140,   &
            '=========',    T150,   &
            '=========',    T160,   &
            '=========',    T170,   &
            '=========',    T180,   &
            '=========',    T190,   &
            '=========',    T200,   &
            '=========',    T210,   &
            '=========',    T220,   &
            '=========',    T230,   &
            '=========',    T240,   &
            '========='             &
                )  
    

202             FORMAT (  1X,                   &
            F7.5,           T10,    &
            F9.5,           T20,    &
            F9.5,           T30,    &
            F9.5,           T40,    &
            F9.5,           T50,    &
            F9.5,           T60,    &
            F9.5,           T70,    &
            F9.5,           T80,    &
            F9.5,           T90,    &
            F9.5,           T100,   &
            F9.5,           T110,   &
            F9.5,           T120,   &
            F9.5,           T130,   &
            F9.5,           T140,   &
            F9.5,           T150,   &
            F9.5,           T160,   &
            F9.5,           T170,   &
            F9.5,           T180,   &
            F9.5,           T190,   &
            F9.5,           T200,   &
            F9.5,           T210,   &
            F9.5,           T220,   &
            F9.5,           T230,   &
            F9.5,           T240,   &
            F9.5                    &
                ) 
    
    ! Format for histogram
203  FORMAT (1X, 'Sudden stops: ', /  &
            'beta   ',      T10,    &
            'theta    ',    T20,    &
            'sigma_l  ',    T30,    &
            'sigma_h  ',    T40,    &
            '  N',          T45,    &
            '   PRE   ',    T55,    &
            '  TRANS  ',    T65,    &
            '   POST  ')  
     
204         FORMAT (1X,                    &
            F7.5,           T10,    &
            F9.5,           T20,    &
            F9.5,           T30,    &
            F9.5,           T40,    &
            I3, T45, F8.5, T55, F8.5, T65, F8.5)
     
    
    OPEN (UNIT=10, FILE="CAL_FLOW.OUT", POSITION="REWIND")
    OPEN (UNIT=12, FILE="HIST_FLOW.OUT", POSITION="REWIND")
    
    ! Header
    WRITE (10,200)
    WRITE (10,201)
    WRITE (12,203)
    
    ! Simulation total
    simul_tot   =   DBLE(N_BETA)*       &
                    DBLE(N_THETA)*      &
                    DBLE(N_SIGMA_L)*    &    
                    DBLE(N_SIGMA_H)
    
    ! Simulation counter
    simul_num   = ZERODP
    
    ! RUN EACH PARAMETER CONFIGURATION
    DO_BETA: DO idx_beta = 1,N_BETA
        DO_THETA: DO idx_theta = 1,N_THETA
            DO_SIGMA_L: DO idx_sigma_l = 1,N_SIGMA_L
                DO_SIGMA_H: DO idx_sigma_h = 1,N_SIGMA_H
                        
                    ! Simulation counter
                    simul_num   = simul_num + ONEDP
                    
                    ! Change parameter space
                    beta        = beta_grid(idx_beta)
                    theta       = theta_grid(idx_theta)
                    sigma_l     = sigma_l_grid(idx_sigma_l)
                    sigma_h     = sigma_h_grid(idx_sigma_h)

301                     FORMAT (1X, '    simul. # ', F9.1,      &
                                '      out of ', F9.1,      &
                            /,1X,                           &
                                ' beta = ', F8.5,      &
                                ' | theta = ', F8.5,      &
                                ' | sigma_l = ', F8.5,      &
                                ' | sigma_h = ', F8.5) 


                    ! PRINT
                    WRITE (*,301)   simul_num,  &
                                    simul_tot,  &
                                    beta,       &
                                    theta,      &
                                    sigma_l,    &
                                    sigma_h                             
                        
                    ! Construct and store PMF distribution
                    CALL createPMF

                    ! Construct and store the feasible set
                    CALL buildFeasible

                    ! Solve for value function
                    CALL iterateVF

                    ! Simulate the model
                    CALL simulation(simul_results)

                    ! Store simulation results
                    RESULTS_GRID(idx_beta,  &
                            idx_theta,      &
                            idx_sigma_l,    &
                            idx_sigma_h)    = simul_results
                        
                        
                    ! print on the go in the file
                    WRITE (10,202)  RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%BETA_VAL,      &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%THETA_VAL,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SIGMA_L_VAL,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SIGMA_H_VAL,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_INI_PRE,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_INI_TRANS,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_INI_POST,    &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_OLD_AVG,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_FREQ_PRE,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_FREQ_TRANS, &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_FREQ_POST,  &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_FREQ_NONTRANS, &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_H_PRE,       &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_H_TRANS,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%R_H_POST,      &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%HAIRCUT_ALL,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%HAIRCUT_COND,  &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%PHI_S_PRE,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%PHI_S_TRANS,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%PHI_S_POST,    &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%Rold_erg_min,    &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%Rold_erg_max,  &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%rho_PRE,       &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%rho_TRANS,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%rho_POST

                    DO idx_hist = 1,N_HIST
                        WRITE(12,204) RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%BETA_VAL,      &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%THETA_VAL,     &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SIGMA_L_VAL,   &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SIGMA_H_VAL,   &
                                    idx_hist-1,                               &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_hist_pre(idx_hist),      &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_hist_trans(idx_hist),    &
                                    RESULTS_GRID(idx_beta,  &
                                        idx_theta,     &
                                        idx_sigma_l,   &
                                        idx_sigma_h)%SS_hist_post(idx_hist)

                    END DO   


                END DO DO_SIGMA_H
            END DO DO_SIGMA_L
        END DO DO_THETA
    END DO DO_BETA                  

    ! Close file
    CLOSE (unit=10)
    CLOSE (unit=12)
    
END SUBROUTINE baseline

!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 FEASIBILITY     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE buildFeasible
	!! This subroutine fills the matrix feasibles
	!! with 0 or 1 indicating whether a given contract 
	!! is feasible and creates initial values
  
	! LOCAL VARIABLES
    INTEGER                                 :: idx_prior
    INTEGER                                 :: idx_rsrv
    INTEGER                                 :: idx_cap
    
    
	! PROGRAM

111  FORMAT (1X, ' Build feasible set ... ') 
	 
	! PRINT
	WRITE (*,111)
        
    DO_RHO: DO idx_prior = 1,N_RHO
		DO_RES: DO idx_rsrv = 1,N_R
			DO_KAP: DO idx_cap = 1,N_K

100  FORMAT (1X, '(i,j,k) = (', 3I3,') || ', &
                ' rho = ', G12.5,' rsrv_pre = ', G12.5,' cap = ', G12.5) 
                  
                
                ! PRINT
                IF (TRACE_TRACK) THEN
                    WRITE (*,100)   idx_prior,  & 
                                    idx_rsrv,   &
                                    idx_cap,    &
                                    rho_grid(idx_prior),    &
                                    r_grid(idx_rsrv),       &
                                    k_grid(idx_cap)  
                END IF                
                    
                ! SOLVE FOR FESASIBILITY AND INTEREST RATE
                CALL solvecontract( & 
                        rho_grid(idx_prior),    &
                        r_grid(idx_rsrv),       &
                        k_grid(idx_cap),        &
                        FSBLR_H(idx_prior,idx_rsrv,idx_cap),    &
                        FSBLMAP(idx_prior,idx_rsrv,idx_cap)     &
                                )
                    
                
                ! FULLY CHARACTERIZE THE CONTRACT IF FEASIBLE
                IS_FSBL: IF ( FSBLMAP(idx_prior,idx_rsrv,idx_cap) ) THEN
                    
                    ! RESERVES INI
                    FSBLRSRV(idx_prior,idx_rsrv,idx_cap) =                  &
                                rsrv_ini(                                   &
                                    r_grid(idx_rsrv),                       &
                                    k_grid(idx_cap)                         &
                                        )
                    
                    ! R_L
                    FSBLR_L(idx_prior,idx_rsrv,idx_cap) =                   &
                                yield_crisis(                               &
                                    r_grid(idx_rsrv),                       &
                                    k_grid(idx_cap),                        &
                                    FSBLR_H(idx_prior,idx_rsrv,idx_cap)     &
                                            )
                    
                    ! PHI_LB
                    FSBLPHI_LB(idx_prior,idx_rsrv,idx_cap) =                &
                                phi_lb(                                     &
                                    r_grid(idx_rsrv),                       &
                                    k_grid(idx_cap),                        &
                                    FSBLR_H(idx_prior,idx_rsrv,idx_cap)     &
                                            )
                    ! PHI_UB
                    FSBLPHI_UB(idx_prior,idx_rsrv,idx_cap) =                &
                                phi_ub(                                     &
                                    r_grid(idx_rsrv),                       &
                                    k_grid(idx_cap),                        &
                                    FSBLR_H(idx_prior,idx_rsrv,idx_cap)     &
                                            ) 
                    ! OUTPUT_CRISIS
                    FSBLY_SS(idx_prior,idx_rsrv,idx_cap)    =               &
                                net_output_crisis(                          &
                                    r_grid(idx_rsrv),                       &
                                    k_grid(idx_cap),                        &
                                    FSBLR_H(idx_prior,idx_rsrv,idx_cap)     &
                                            )
     
101  FORMAT (1X, ' >>> ', ' feas = ', I2, / , &
             1X, ' >>> ', ' rsrv_ini = ', G15.8,' 1 + r_H = ', G15.8, ' 1 + r_L = ', G15.8, / , &
             1X, ' >>> ', ' phi_lb   = ', G15.8,' phi_ub  = ', G15.8, ' y_ss    = ', G15.8,/)                                        
                    
                    ! PRINT
                    IF (TRACE_TRACK) THEN
                        WRITE (*,101)                                       &   
                                    FSBLMAP(idx_prior,idx_rsrv,idx_cap),    &   
                                    FSBLRSRV(idx_prior,idx_rsrv,idx_cap),   &
                                    FSBLR_H(idx_prior,idx_rsrv,idx_cap),    &
                                    FSBLR_L(idx_prior,idx_rsrv,idx_cap),    &                        
                                    FSBLPHI_LB(idx_prior,idx_rsrv,idx_cap), &   
                                    FSBLPHI_UB(idx_prior,idx_rsrv,idx_cap), &
                                    FSBLY_SS(idx_prior,idx_rsrv,idx_cap)
                    END IF
                END IF IS_FSBL

            END DO DO_KAP

            ! INDICATE WHETHER THERE IS ANY FEASIBLE SUBCONTRACT
            FSBLANY(idx_prior,idx_rsrv) =  MAXVAL(FSBLMAP(idx_prior,idx_rsrv,:))
            
103         FORMAT (1X, '(i,j) = (', 2I3,') || ', &
                ' rho = ', G12.5,' rsrv_pre = ', G12.5,' any feasible contract = ', I2)                                      

            ! PRINT
            IF (TRACE_TRACK) THEN
                WRITE (*,103) idx_prior,                &
                idx_rsrv,                               & 
                rho_grid(idx_prior),                    &
                r_grid(idx_rsrv),                       &  
                FSBLANY(idx_prior,idx_rsrv)
            END IF            

        END DO DO_RES

        ! INDICATE MINIMUM RESERVES WITH FEASIBLE SUBCONTRACT
        IF ( ANY(FSBLANY(idx_prior,:) .EQ. 1) ) THEN
            FSBLMIN(idx_prior)  = r_grid(MINVAL(MAXLOC(FSBLANY(idx_prior,:))))
        ELSE
            ! set to RMAX+EPS to indicate that no feasible point on grid
            FSBLMIN(idx_prior)  = RMAX + EPS
        END IF
        

104  FORMAT (1X, '(i) = (', I3,') || ', &
                ' rho = ', G12.5,' min feasible rsrv_pre = ', G12.5)                                      

            ! PRINT
            IF (trace_code) THEN
                WRITE (*,104) idx_prior,                &
                rho_grid(idx_prior),                    & 
                FSBLMIN(idx_prior)
            END IF          
        
    END DO DO_RHO

!112  FORMAT (/,1X, ' BUILD FEASIBLE SET - DONE ') 
!	 
!	! PRINT
!	WRITE (*,112)
	
END SUBROUTINE buildFeasible

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!! PRINT CALIBRATION AND SIMULATION RESULTS !!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE printResults

    ! LOCAL VARIABLES
    INTEGER                                 :: idx_beta
    INTEGER                                 :: idx_theta
    INTEGER                                 :: idx_sigma_l
    INTEGER                                 :: idx_sigma_h
    
100 FORMAT (  1X,                   &
            'beta   ',      T10,    &
            'theta    ',    T20,    &
            'sigma_l  ',    T30,    &
            'sigma_h  ',    T40,    &
            'res pre  ',    T50,    &
            'res tra  ',    T60,    &
            'res pos  ',    T70,    &
            'res sav  ',    T80,    &
            'ss % pre ',    T90,   &
            'ss % tra ',    T100,   &
            'ss % pos ',    T110,   &
            'ss % nont',    T120,   &
            'R_H pre  ',    T130,   &
            'R_H tra  ',    T140,   &
            'R_H pos  ',    T150,   &
            'hcut all ',    T160,   &
            'hcut con ',    T170,   &
            'phiS pre ',    T180,   &
            'phiS tra ',    T190,   &
            'phiS pos ',    T200,   &
            'MIN erR_0',    T210,   &
            'MAX erR_0',    T220,   &
            'rho pre  ',    T230,   &
            'rho trans',    T240,   &
            'rho post '             &
                )
    
101             FORMAT (  1X,                   &
            '=======',      T10,    &
            '=========',    T20,    &
            '=========',    T30,    &
            '=========',    T40,    &
            '=========',    T50,    &
            '=========',    T60,    &
            '=========',    T70,    &
            '=========',    T80,    &
            '=========',    T90,    &
            '=========',    T100,   &
            '=========',    T110,   &
            '=========',    T120,   &
            '=========',    T130,   &
            '=========',    T140,   &
            '=========',    T150,   &
            '=========',    T160,   &
            '=========',    T170,   &
            '=========',    T180,   &
            '=========',    T190,   &
            '=========',    T200,   &
            '=========',    T210,   &
            '=========',    T220,   &
            '=========',    T230,   &
            '=========',    T240,   &
            '========='             &
                )  
    

102             FORMAT (  1X,                   &
            F7.5,           T10,    &
            F9.5,           T20,    &
            F9.5,           T30,    &
            F9.5,           T40,    &
            F9.5,           T50,    &
            F9.5,           T60,    &
            F9.5,           T70,    &
            F9.5,           T80,    &
            F9.5,           T90,    &
            F9.5,           T100,   &
            F9.5,           T110,   &
            F9.5,           T120,   &
            F9.5,           T130,   &
            F9.5,           T140,   &
            F9.5,           T150,   &
            F9.5,           T160,   &
            F9.5,           T170,   &
            F9.5,           T180,   &
            F9.5,           T190,   &
            F9.5,           T200,   &
            F9.5,           T210,   &
            F9.5,           T220,   &
            F9.5,           T230,   &
            F9.5,           T240,   &
            F9.5                    &
                )  
    
    OPEN (unit=11, file="CAL_FINAL.OUT", position="rewind")
    
    WRITE (11,100)
    WRITE (11,101)
    
    ! PRINT EACH PARAMETER CONFIGURATION
    DO_BETA: DO idx_beta = 1,N_BETA
        DO_THETA: DO idx_theta = 1,N_THETA
            DO_SIGMA_L: DO idx_sigma_l = 1,N_SIGMA_L
                DO_SIGMA_H: DO idx_sigma_h = 1,N_SIGMA_H
                    
                    WRITE (11,102)      RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%BETA_VAL,      &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%THETA_VAL,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SIGMA_L_VAL,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SIGMA_H_VAL,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_INI_PRE,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_INI_TRANS,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_INI_POST,    &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_OLD_AVG,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SS_FREQ_PRE,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SS_FREQ_TRANS, &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SS_FREQ_POST,  &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%SS_FREQ_NONTRANS, &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_H_PRE,       &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_H_TRANS,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%R_H_POST,      &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%HAIRCUT_ALL,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%HAIRCUT_COND,  &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%PHI_S_PRE,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%PHI_S_TRANS,   &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%PHI_S_POST,    &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%Rold_erg_min,    &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%Rold_erg_max,  &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%rho_PRE,       &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%rho_TRANS,     &
                                        RESULTS_GRID(idx_beta,  &
                                            idx_theta,     &
                                            idx_sigma_l,   &
                                            idx_sigma_h)%rho_POST
    
                END DO DO_SIGMA_H
            END DO DO_SIGMA_L
        END DO DO_THETA
    END DO DO_BETA
      
    CLOSE (unit=11)


END SUBROUTINE printResults


END MODULE AUXILIARY