MODULE toolkit

!!!!!!!!!!!!!!!!!!!!!!!!
!!   USE STATEMENTS   !!
!!!!!!!!!!!!!!!!!!!!!!!! 

USE PARAMETERS
USE GLOBAL

IMPLICIT NONE 

CONTAINS

!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 UTILITY      !!
!!!!!!!!!!!!!!!!!!!!!!!! 

ELEMENTAL FUNCTION utility(c) !! utility FUNCTION
	!! Simple linear utility FUNCTION

	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: c
    REAL(DP)                :: utility
    
	! PROGRAM
    utility	=	c
    
END FUNCTION utility

!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!    	PARETO CDF 		  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

ELEMENTAL FUNCTION cdf_sigma_L(x)
	!! Pareto bounded CDF with sigma = sigma_low
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: cdf_sigma_L
    
	! PROGRAM
    cdf_sigma_L	=	1.0_DP-(1.0_DP-x)**(1.0_DP/sigma_L)
    
END FUNCTION cdf_sigma_L

ELEMENTAL FUNCTION cdf_sigma_H(x)
	!! Pareto bounded CDF with sigma = sigma_high
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: cdf_sigma_H
    
	! PROGRAM
    cdf_sigma_H	=	1.0_DP-(1.0_DP-x)**(1.0_DP/sigma_H)
    
END FUNCTION cdf_sigma_H

ELEMENTAL FUNCTION cdf_h(x,rho)
	!! Mixed pareto bounded CDF with prior = rho
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x, rho
    REAL(DP)                :: cdf_h
    
	! PROGRAM
    cdf_h	= 	rho*cdf_sigma_L(x) + (1.0_DP-rho)*cdf_sigma_H(x)
    
END FUNCTION cdf_h

!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!    	PARETO PDF 		  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

ELEMENTAL FUNCTION pdf_sigma_L(x)
	!! Pareto bounded PDF with sigma = sigma_low
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: pdf_sigma_L
    
	! PROGRAM
    pdf_sigma_L	=	(1.0_DP/sigma_L)*((1.0_DP - x)**(-1.0_DP + 1.0_DP/sigma_L))
    
END FUNCTION pdf_sigma_L

ELEMENTAL FUNCTION pdf_sigma_H(x)
	!! Pareto bounded PDF with sigma = sigma_high
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: pdf_sigma_H
    
	! PROGRAM
    pdf_sigma_H	=	(1.0_DP/sigma_H)*((1.0_DP - x)**(-1.0_DP + 1.0_DP/sigma_H))
    
END FUNCTION pdf_sigma_H

ELEMENTAL FUNCTION pdf_h(x,rho)
	!! Mixed pareto bounded PDF with prior = rho
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x, rho
    REAL(DP)                :: pdf_h
    
	! PROGRAM
    pdf_h	=	rho*pdf_sigma_L(x) + (1.0_DP-rho)*pdf_sigma_H(x)
    
END FUNCTION pdf_h

!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   PARETO CDF INTEGRAL  !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!! 

ELEMENTAL FUNCTION icdf_sigma_L(x)
	!! Pareto bounded CDF integral of phi*pdf_sigma(phi) with sigma = sigma_low
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: icdf_sigma_L
    
	! PROGRAM
    icdf_sigma_L	=	(sigma_L-((1.0_DP-x)**(1.0_DP/sigma_L))*(x+sigma_L))/(sigma_L+1.0_DP)
    
END FUNCTION icdf_sigma_L

ELEMENTAL FUNCTION icdf_sigma_H(x)
	!! Pareto bounded CDF integral of phi*pdf_sigma(phi) with sigma = sigma_high
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x
    REAL(DP)                :: icdf_sigma_H
    
	! PROGRAM
    icdf_sigma_H	=	(sigma_H-((1.0_DP-x)**(1.0_DP/sigma_H))*(x+sigma_H))/(sigma_H+1.0_DP)
    
END FUNCTION icdf_sigma_H

ELEMENTAL FUNCTION icdf_h(x,rho)
	!! Mixed pareto bounded CDF inegral with prior = rho
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x, rho
    REAL(DP)                :: icdf_h
    
	! PROGRAM
    icdf_h	= rho*icdf_sigma_L(x) + (1.0_DP-rho)*icdf_sigma_H(x)
    
END FUNCTION icdf_h

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 BELIEF TRANSITION      !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

ELEMENTAL FUNCTION m(x,phi,rho)
	!! m function where 
    !! x is a posterior value, phi a shock, rho a prior.
    !! m gives the corresponding product of shocks
    !! See appendix for formula
    !! addded min of (-eps, m) to fix negative m issue at x values close to one
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    ::  x               ! posterior belief value
    REAL(DP), INTENT(IN)    ::  phi             ! shock realized in a country
    REAL(DP), INTENT(IN)    ::  rho             ! prior belief
    REAL(DP)                ::  m               ! product of the shocks in other countries
    
	! PROGRAM
    m	= 	MIN(-eps, -LOG(1.0_DP-phi)	&
			+  1.0_DP/( 1.0_DP/sigma_H - 1.0_DP/sigma_L )	&
			*  LOG( ( 1.0_DP/x - 1.0_DP )*(rho/(1.0_DP-rho))*(sigma_H/sigma_L)**(N_eme) ) )
    
END FUNCTION m

ELEMENTAL FUNCTION mprime(x)
	!! Derivative of the function m(x,phi,rho) wrt x
    !! See appendix for formula
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    ::  x               ! posterior belief value
    REAL(DP)                ::  mprime
    
	! PROGRAM
    mprime	= 	-1.0_DP/( 1.0_DP/sigma_H - 1.0_DP/sigma_L )/(x*(1.0_DP-x))
    
END FUNCTION mprime

ELEMENTAL FUNCTION realfactorial(n)
	!! Computes the factorial of n, using real numbers
	
	! INPUT & OUTPUT
    INTEGER, INTENT(IN)     ::  n
    
    ! LOCAL
    INTEGER                 ::  i
    REAL(DP)                ::  aux
    REAL(DP)                ::  realfactorial
    
	! PROGRAM
    aux = 1.0_DP
    IF (n>=2) THEN
		DO i = 2,n
			aux =   aux*DBLE(i)
		END DO
    ENDIF
    realfactorial = aux
    
END FUNCTION realfactorial

ELEMENTAL FUNCTION afunc(y,n,sigma)
	!! "a" function
    !! This is the main function used to derive the cdf of 
    !! the log-sum of shocks realized across countries
    
    !! See appendix for recursive formula
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: y                ! value of log-sum of shocks
    REAL(DP), INTENT(IN)    :: sigma            ! Pareto parameter (regime)
	INTEGER, INTENT(IN)     :: n                ! number of draws summed (countries)
    
    REAL(DP)                :: afunc
    
    ! LOCAL
    INTEGER	                :: k
    REAL(DP)                :: facto, aux
    
	! PROGRAM
	aux = sigma
	IF (n>=2) THEN
		DO k = 2,n
			aux	=	aux + DBLE((-1)**(k-1))/( realfactorial(k-1)*sigma**(k-2) )*y**(k-1)
		END DO
    ENDIF
    afunc = aux
	
END FUNCTION afunc

ELEMENTAL FUNCTION alta(y,n,sigma)
	!! alternative "A" function
    !! A = (1/sigma) * a * exp(y/sigma)
	
    !! This is the main function used to derive the cdf of 
    !! the log-sum of shocks realized across countries
    
    !! See appendix for recursive formula
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: y                ! value of log-sum of shocks
    REAL(DP), INTENT(IN)    :: sigma            ! Pareto parameter (regime)
	INTEGER, INTENT(IN)     :: n                ! number of draws summed (countries)
    
    REAL(DP)	            :: alta
    
    ! LOCAL
    INTEGER	                :: k
    REAL(DP)	            :: facto, aux
    
	! PROGRAM
	aux = sigma/sigma*EXP(y/sigma)
	IF (n>=2) THEN
		DO k = 2,n
			aux	=	aux + DBLE((-1)**(k-1))/( realfactorial(k-1)*sigma**(k-2) )*y**(k-1)/sigma*EXP(y/sigma)
		END DO
    ENDIF
	alta = aux
    
END FUNCTION alta

ELEMENTAL FUNCTION afuncprime(x,n,sigma)
	!! Derivative of "a" function wrt x
    !! The "a" function is the main function used to derive the cdf of 
    !! the log-sum of shocks realized across countries
    
    !! See appendix for recursive formula    
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x                ! value of log-sum of shocks
    REAL(DP), INTENT(IN)    :: sigma            ! Pareto parameter (regime)
	INTEGER, INTENT(IN)     :: n                ! number of draws summed (countries)
    
    REAL(DP)                :: afuncprime
    
    ! LOCAL    
    INTEGER	                :: k
    REAL(DP)                :: facto
    
	! PROGRAM
	afuncprime = 0.0_DP
	IF (n>=2) THEN
		DO k = 2,n
			afuncprime	=	afuncprime + DBLE((-1)**(k-1))/( realfactorial(k-2)*sigma**(k-2) )*x**(k-2)
		END DO
    ENDIF
	
END FUNCTION afuncprime

ELEMENTAL FUNCTION capAfunc(x,n,sigma)    
    !! A = (1/sigma) * a * exp(x/sigma)	
    !! This is the main function used to derive the cdf of 
    !! the log-sum of shocks realized across countries    
    !! See appendix for formula and derivation
    !! note: this is exactly the same as "alta", with y not x.
    	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x                ! value of log-sum of shocks
    REAL(DP), INTENT(IN)    :: sigma            ! Pareto parameter (regime)
	INTEGER, INTENT(IN)     :: n                ! number of draws summed (countries)
    
    REAL(DP)	            :: capAfunc
    
	! PROGRAM
	capAfunc        = EXP(x/sigma)*afunc(x,n,sigma)/sigma
	
END FUNCTION capAfunc

ELEMENTAL FUNCTION capAprime(x,n,sigma)
    !! Derivative of "A" function wrt x
    !! The "A" function is the main function used to derive the cdf of 
    !! the log-sum of shocks realized across countries
    
    !! See appendix for formula    
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: x                ! value of log-sum of shocks
    REAL(DP), INTENT(IN)    :: sigma            ! Pareto parameter (regime)
	INTEGER, INTENT(IN)     :: n                ! number of draws summed (countries)
    
    REAL(DP)                :: capAprime
    
	! PROGRAM
	capAprime       = EXP(x/sigma)*( afuncprime(x,n,sigma) + afunc(x,n,sigma)/sigma )/sigma
    capAprime       = MAX(0.0_DP,capAprime)
	
END FUNCTION capAprime

ELEMENTAL FUNCTION xbar(phi,rho)
	!! Upper bound of posterior 
    !! See appendix for formula 
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: phi              ! realized shock
    REAL(DP), INTENT(IN)    :: rho              ! prior
    REAL(DP)                :: xbar
    
	! PROGRAM
	xbar = 1.0_DP/( 1.0_DP + (1.0_DP-phi)**( 1.0_DP/sigma_H - 1.0_DP/sigma_L )  &
                                *((1.0_DP-rho)/rho)*(sigma_L/sigma_H)**N_eme )
	
END FUNCTION xbar

ELEMENTAL FUNCTION g(x,phi,rho)
	!! PDF of posteriors given prior and local shock - see appendix
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)	:: x            ! posterior
    REAL(DP), INTENT(IN)	:: phi          ! local shock in one country
    REAL(DP), INTENT(IN)	:: rho          ! prior
    REAL(DP)                :: g
    
	! PROGRAM
    IF (x<=xbar(phi,rho)) THEN
        g	= 	rho*( capAprime(m(x,phi,rho),N_eme-1,sigma_L)*mprime(x) )	&
			+ (1.0_DP-rho)*( capAprime(m(x,phi,rho),N_eme-1,sigma_H)*mprime(x) )
    ELSE
        g   =   0.0_DP
    ENDIF
    
END FUNCTION g

ELEMENTAL FUNCTION altg(x,phi,rho)
    !! Approximated PMF of posteriors on RHO_P GRID
    !! See appendix for technical notes
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)	:: x            ! posterior
    REAL(DP), INTENT(IN)	:: phi          ! local shock in one country
    REAL(DP), INTENT(IN)	:: rho          ! prior
    REAL(DP)                :: altg
    
	! PROGRAM
    IF ( ( x > xbar(phi,rho) ) .AND. ( x > rhopmin ) .AND. ( x < rhopmax ) ) THEN
        altg    =   0.0_DP
    ELSEIF (x<=rhopmin) THEN
        altg    = 2.0_DP/(rhop_grid(2)-rhop_grid(1))*bigG(rhop_grid(2),phi,rho) - g(rhop_grid(2),phi,rho)
    ELSEIF (x>=rhopmax) THEN
        altg    = 2.0_DP/(rhop_grid(N_rhop)-rhop_grid(N_rhop-1))*( 1 - bigG(rhop_grid(N_rhop-1),phi,rho) ) - g(rhop_grid(N_rhop-1),phi,rho)
    ELSE
        altg = g(x,phi,rho)
    ENDIF
    
END FUNCTION altg

ELEMENTAL FUNCTION bigG(x,phi,rho)
	!! CDF of posteriors given prior and local shock
    !! See appendix
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)	:: x            ! posterior
    REAL(DP), INTENT(IN)	:: phi          ! local shock in one country
    REAL(DP), INTENT(IN)	:: rho          ! prior
    
    REAL(DP)                :: bigG
    
	! PROGRAM
    IF (x<=xbar(phi,rho)) THEN
        bigG	= 	rho*capAfunc( m(x,phi,rho), N_eme-1, sigma_L  )   &
                +   (1.0_DP-rho)*capAfunc( m(x,phi,rho), N_eme-1, sigma_H  )
    ELSE
        bigG   =   1.0_DP
    ENDIF
    
END FUNCTION bigG

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 OTHER FUNCTIONS      !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

FUNCTION prodvec(x,N)
	!! product of elements in vector of DIMENSION N
	
	! INPUT & OUTPUT
    REAL(DP), DIMENSION(N)  :: x
    INTEGER, INTENT (IN)    :: N
    INTEGER                 :: i
    REAL(DP)                :: prodvec
    
    prodvec = 1.0_DP
	! PROGRAM
    DO i = 1,N
        prodvec = prodvec*x(i)
    END DO
        
END FUNCTION prodvec

FUNCTION linspace(a,b,N)
	!! linspace

	! INPUT & OUTPUT
	INTEGER, INTENT(IN)         :: N
	REAL(DP), INTENT(IN)        :: a,b
    
    ! OUTPUT
    REAL(DP), DIMENSION(N)      :: linspace
    
    ! LOCAL
	REAL(DP)                    :: delta
	INTEGER                     :: i

	! PROGRAM
	delta           = (b-a)/DBLE(N-1)
	linspace        = a + delta*(/(DBLE(i),i=0,N-1)/)
	
END FUNCTION linspace

FUNCTION trapz(X,Y,siz)
	!! computes the integration of Y with respect to X using trapezoidal integration. 
	!! X and Y must be one-DIMENSIONal arrays with the same DIMENSION siz

	! INPUT & OUTPUT
	INTEGER, INTENT(IN) 				    :: siz
	REAL(DP), DIMENSION(siz), INTENT(IN)	:: X, Y
    
    ! LOCAL VARIABLES
	REAL(DP)                				:: ans, trapz
	INTEGER                            	    :: i

	! PROGRAM
	ans = 0.0_DP
	DO i = 1, siz-1
		ans = ans + 0.5_DP*( Y(i+1) + Y(i) )*( X(i+1) - X(i) )
    END DO
	trapz = ans
	
END FUNCTION trapz

FUNCTION interp1(X,Y,siz,xi)
	!! Interpolates to find the values of the underlying function Y at the points in the vector or array xi.
	!! X and Y must be one-DIMENSIONal arrays with the same DIMENSION siz
	!! xi must be a scalar
	!! if xi exceeds the bounds of X, i.e. requiring extrapolation, then Yi is set to the closest point
	
	! INPUT & OUTPUT
    INTEGER, INTENT(IN) 					:: siz
    REAL(DP), DIMENSION(siz), INTENT(IN)	:: X, Y
    REAL(DP), INTENT(IN)	                :: xi
    REAL(DP)                                :: interp1
    INTEGER								    :: ind
    
	! PROGRAM
	ind = MINLOC( ABS( xi - X ) , 1 )
	outer: IF ( xi >= X(ind) ) THEN
		inner1: IF  ( ind == siz ) THEN
			interp1 = Y(ind)
		ELSE
			interp1 =	&
				( Y(ind+1)*(xi-X(ind)) + Y(ind)*(X(ind+1)-xi) )	&
				/( X(ind+1) - X(ind) )
		END IF inner1
    ELSE ! IF ( xi < X(ind) )
		inner2: IF  ( ind == 1 ) THEN
			interp1 = Y(ind)
		ELSE
			interp1 =	&
				( Y(ind)*(xi-X(ind-1)) + Y(ind-1)*(X(ind)-xi) )	&
				/( X(ind) - X(ind-1) )
		END IF inner2
    END IF outer
	
END FUNCTION interp1

FUNCTION interp2(X1,X2,Y,siz1,siz2,xi1,xi2)
	!! Interpolates to find the values of the underlying function Y at the point (xi1,xi2).
	!! X1 (X2) must be one-DIMENSIONal arrays with DIMENSION siz1 (siz2)
    !! Y must be a two-DIMENSIONal array of DIMENSION siz1 x siz2
	!! xi1, xi2 must be scalars
	!! if xi exceeds the bounds of X, i.e. requiring extrapolation, then Yi is set to the closest point
	
	! INPUT & OUTPUT
    INTEGER, INTENT(IN) 					    :: siz1, siz2
    REAL(DP), DIMENSION(siz1), INTENT(IN)	    :: X1
    REAL(DP), DIMENSION(siz2), INTENT(IN)	    :: X2
    REAL(DP), DIMENSION(siz1,siz2), INTENT(IN)	:: Y
    REAL(DP), INTENT(IN)	                    :: xi1 ,xi2
    REAL(DP)                                    :: Yind, Yindp, Yindm
    REAL(DP)                                    :: interp2
    INTEGER								        :: ind
    
    ! PROGRAM
	ind = MINLOC( ABS( xi1 - X1 ) , 1 )
    outer: IF ( xi1 >= X1(ind) ) THEN
		inner1: IF  ( ind == siz1 ) THEN
			interp2 = INTERP1(X2,Y(ind,:),siz2,xi2)
        ELSE
            Yind = INTERP1(X2,Y(ind,:),siz2,xi2)
            Yindp = INTERP1(X2,Y(ind+1,:),siz2,xi2)
            interp2 = &
                ( Yindp*(xi1-X1(ind)) + Yind*(X1(ind+1)-xi1) )	&
				/( X1(ind+1) - X1(ind) )
		END IF inner1
    ELSE ! IF ( xi1 < X1(ind) )
		inner2: IF  ( ind == 1 ) THEN
			interp2 = INTERP1(X2,Y(ind,:),siz2,xi2)
        ELSE
            Yind = INTERP1(X2,Y(ind,:),siz2,xi2)
            Yindm = INTERP1(X2,Y(ind-1,:),siz2,xi2)
			interp2 =	&
				( Yind*(xi1-X1(ind-1)) + Yindm*(X1(ind)-xi1) )	&
				/( X1(ind) - X1(ind-1) )
		END IF inner2
    END IF outer
	
END FUNCTION interp2

FUNCTION interp3(X1,X2,X3,Y,siz1,siz2,siz3,xi1,xi2,xi3)
	!! Interpolates to find the values of the underlying function Y at the point (xi1,xi2,xi3).
	!! X1,X2,X3 must be one-DIMENSIONal arrays with DIMENSION siz1,siz2,siz3
    !! Y must be a three-DIMENSIONal array of DIMENSION siz1 x siz2 x siz3
	!! xi1, xi2, xi3 must be scalars
	!! if xi exceeds the bounds of X, i.e. requiring extrapolation, then Yi is set to the closest point
	
	! INPUT & OUTPUT
    INTEGER, INTENT(IN) 					        :: siz1, siz2, siz3
    REAL(DP), DIMENSION(siz1), INTENT(IN)	        :: X1
    REAL(DP), DIMENSION(siz2), INTENT(IN)	        :: X2
    REAL(DP), DIMENSION(siz3), INTENT(IN)	        :: X3
    REAL(DP), DIMENSION(siz1,siz2,siz3), INTENT(IN)	:: Y
    REAL(DP), INTENT(IN)	                        :: xi1 ,xi2, xi3
    REAL(DP)                                        :: Yind, Yindp, Yindm
    REAL(DP)                                        :: interp3
    INTEGER								            :: ind
    
    ! PROGRAM
	ind = MINLOC( ABS( xi1 - X1 ) , 1 )
    outer: IF ( xi1 >= X1(ind) ) THEN
		inner1: IF  ( ind == siz1 ) THEN
			interp3 = INTERP2(X2,X3,Y(ind,:,:),siz2,siz3,xi2,xi3)
        ELSE
            Yind = INTERP2(X2,X3,Y(ind,:,:),siz2,siz3,xi2,xi3)
            Yindp = INTERP2(X2,X3,Y(ind+1,:,:),siz2,siz3,xi2,xi3)
            interp3 = &
                ( Yindp*(xi1-X1(ind)) + Yind*(X1(ind+1)-xi1) )	&
				/( X1(ind+1) - X1(ind) )
		END IF inner1
    ELSE ! IF ( xi1 < X1(ind) )
		inner2: IF  ( ind == 1 ) THEN
			interp3 = INTERP2(X2,X3,Y(ind,:,:),siz2,siz3,xi2,xi3)
        ELSE
            Yind = INTERP2(X2,X3,Y(ind,:,:),siz2,siz3,xi2,xi3)
            Yindm = INTERP2(X2,X3,Y(ind-1,:,:),siz2,siz3,xi2,xi3)
			interp3 =	&
				( Yind*(xi1-X1(ind-1)) + Yindm*(X1(ind)-xi1) )	&
				/( X1(ind) - X1(ind-1) )
		END IF inner2
    END IF outer
	
END FUNCTION interp3

RECURSIVE PURE SUBROUTINE bisection(x_min,x_max,func,fzero,fval,errcode)
	!! This SUBROUTINE solves for the zero 
	!! of the function func
	!! between x_min and x_max
	!! and the value at fval = func(fzero)
    !! An error code is stored in errcode (0=ok, 1=error)
	
	! INPUT & OUTPUT
	REAL(DP), INTENT(IN) 					:: x_min, x_max
	!REAL(DP), EXTERNAL                      :: func
    INTERFACE
        REAL(SELECTED_REAL_KIND(p=13)) ELEMENTAL FUNCTION func(x)
            REAL(SELECTED_REAL_KIND(p=13)), INTENT(IN)            :: x
        END FUNCTION
    END INTERFACE
    
	REAL(DP), INTENT(OUT) 	                :: fzero
    REAL(DP), INTENT(OUT) 	                :: fval
    INTEGER, INTENT(OUT) 					:: errcode
	
	! LOCAL VARIABLES
	REAL(DP)                                :: f_low, f_high, x_low, x_high, x_mid, f_mid, x_diff 
    REAL(DP)                                :: fv, fz
    INTEGER                                 :: errc
     
	! PROGRAM
    
    errcode     = 0
    x_low       = x_min
    f_low       = func(x_low)
    x_high      = x_max
    f_high      = func(x_high)
    
    x_mid       = (x_min + x_max)/2.0_DP
    f_mid       = func(x_mid)
    
    x_diff      = x_low - x_high
    
    IF (abs(f_mid) < EPS) THEN
        errcode = 0
        fval    = f_mid
        fzero   = x_mid
        
    ELSE IF (abs(x_diff) < EPS) THEN
        errcode = 1
        fval    = f_mid
        fzero   = x_mid        
        
    ELSE IF (f_mid*f_low < ZERODP) THEN
        CALL bisection(x_low,x_mid,func,fz,fv,errc)
        fzero   = fz
        fval    = fv
        errcode = errc
        
    ELSE IF (f_mid*f_high < ZERODP) THEN
        CALL bisection(x_mid,x_high,func,fz,fv,errc)
        fzero   = fz
        fval    = fv
        errcode = errc
        
    ELSE IF (f_high*f_low > ZERODP) THEN
        errcode = 1
        fzero   = x_low
        fval    = f_low        
    END IF	
														
END SUBROUTINE bisection
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!   	 GRID CONSTRUCTION      !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

SUBROUTINE creategrid(grid, siz, grid_min, grid_pivot, grid_max, curvature)
	!! This SUBROUTINE fills the array grid 
	!! with siz points
	!! from grid_min to grid_max
	!! and a curvature around grid_pivot
	
	! INPUT & OUTPUT
    INTEGER, INTENT(IN) 					:: siz
	REAL(DP), INTENT(IN) 					:: grid_min, grid_max, grid_pivot
	REAL(DP), INTENT(IN) 					:: curvature
    
	REAL(DP), DIMENSION(SIZ), INTENT(INOUT) 	:: grid
	
	! LOCAL VARIABLES
	INTEGER                                 :: i, N_med
     
	! PROGRAM
    
    
    IF_SINGLE: IF (siz .EQ. 1) THEN        
        ! if size of the grid is one
        ! then set to lower bound
        grid        =   grid_min
    ELSE

        N_med       = MAX(1,CEILING(siz*((grid_pivot - grid_min)/ &
                            (grid_max - grid_min))))

        IF (N_med .GT. 1) THEN
            FORALL (i=1:N_med) grid(i) = DBLE(i-1)/DBLE(N_med-1)  	! [0,1] grid evenly split
            grid(1:N_med)	= 1.0_DP - grid(1:N_med)                ! [1,0] grid evenly split
            grid(1:N_med)	= 1.0_DP - grid(1:N_med)**curvature    	! [0,1] grid clustered at 1
            FORALL (i=1:N_med) grid(i) = grid_min + & 
            (grid_pivot-grid_min)*grid(i)   ! grid_min to grid_pivot, clustered
        END IF

        IF (N_med .LT. siz) THEN
            FORALL (i=N_med:siz) grid(i) = DBLE(i-N_med)/DBLE(siz-N_med)   ! even grid between 0 and 1
            grid(N_med:siz)	= grid(N_med:siz)**curvature            ! [0,1] grid clustered at 0
            FORALL (i=N_med:siz) grid(i) = grid_pivot + & 
            (grid_max-grid_pivot)*grid(i)   ! grid_min to grid_pivot, clustered
        END IF

    END IF IF_SINGLE
														
END SUBROUTINE creategrid

FUNCTION rhopgfunc(phi,rho)

    !! This generates a grid of posterior beliefs
    !! between 0 and the highest posterior possible
    !! The spacing is even when the upper bound is low
    !! and unevenly clustered near .5 when the upper bound is high enough

	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)	        :: phi          ! shock
    REAL(DP), INTENT(IN)	        :: rho          ! posterior
    
    REAL(DP), DIMENSION(N_rhop)     :: rhopgfunc
    
    ! LOCAL VARIABLES
    REAL(DP)                        :: xmax
    
    
    rhopgfunc = 0.0_DP
    
    xmax = xbar(phi,rho)
    IF (xmax<=rhopmin) THEN
        rhopgfunc = linspace( xmax*1.0D-10 , rhopmin , N_rhop )
    ELSE
        rhopgfunc(1:INT(N_rhop/2)) = linspace( rhopmin**(1.0_DP/grdpwr), 0.5_DP**(1.0_DP/grdpwr) , INT(N_rhop/2) )
        rhopgfunc = rhopgfunc**grdpwr
        rhopgfunc(INT(N_rhop/2):N_rhop) = linspace( 0.5_DP**grdpwr , rhopmax**grdpwr , N_rhop-INT(N_rhop/2)+1 )
        rhopgfunc(INT(N_rhop/2):N_rhop) = rhopgfunc(INT(N_rhop/2):N_rhop)**(1.0_DP/grdpwr)
    END IF
    
    END FUNCTION rhopgfunc

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  	AUXILIARY FUNCTIONS     !!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    

ELEMENTAL SUBROUTINE solvecontract(prior, rsrv_pre, capital, yield, feasible)
	!! DESCRIPTION
	!! This subroutine derives the feasibility
	!! and the implied time-0 properties of a contract (interest rate)
	!! given: prior, incoming reserves, capital allocation
	
	!! INPUTS
	REAL(DP), INTENT(IN)					:: prior	! Prior belief
    REAL(DP), INTENT(IN)					:: rsrv_pre	! Incoming reserve
	REAL(DP), INTENT(IN)					:: capital	! Capital allocation
	
	!! OUTPUTS
	REAL(DP), INTENT(INOUT)					:: yield	! Promised interest rate 
	INTEGER, INTENT(INOUT)					:: feasible	! Whether the contract is feasible   
	
	!! LOCAL VARIABLES
    INTEGER                                 :: i        ! index
    REAL(DP)                                :: r_lbound ! lower bound for yield
    REAL(DP)                                :: r_ubound ! upper bound for yield
    REAL(DP), DIMENSION(N_INTS+1)           :: r_sample ! sampled values of yield
    REAL(DP), DIMENSION(N_INTS)             :: fvals    ! value of PC at candidate zero
    REAL(DP), DIMENSION(N_INTS)             :: zeros    ! candidate value for PC = 0
    INTEGER, DIMENSION(N_INTS)              :: errcodes ! error code for bisection
    
	!! PROGRAM
    r_lbound            = R
    r_ubound            = MAX( A*capital + rsrv_ini(rsrv_pre, capital) , RHMAX )
    
    FORALL ( i = 1:N_INTS+1 )   
        r_sample(i)     = r_lbound + (r_ubound - r_lbound)*DBLE(i-1)/DBLE(N_INTS)
    END FORALL
      
    DO i = 1,N_INTS               
        CALL bisection(r_sample(i),r_sample(i+1),pc,zeros(i),fvals(i),errcodes(i))   
    END DO
    
    feasible            = 1 - MINVAL( errcodes )
    
    IF ( feasible .EQ. 1) THEN
        yield           = MINVAL( zeros, MASK = errcodes .EQ. 0 )
    ELSE
        yield           = R
    END IF
    
    
CONTAINS

    ELEMENTAL FUNCTION pc(interest)
        !! DESCRIPTION
	    !! This subroutine derives the participation constraint
        !! [expected return given interest rate ] - [world interest rate]
        
        ! INPUT & OUTPUT
    	REAL(DP), INTENT(IN)				:: interest
        
        REAL(DP)                            :: pc
        
        !! PROGRAM        
        pc              = participation(prior, rsrv_pre, capital, interest)
    
    END FUNCTION pc

	
END SUBROUTINE solvecontract

ELEMENTAL FUNCTION participation(prior, rsrv_pre, capital, yield)
    !! DESCRIPTION
	!! This subroutine derives the participation constraint
    !! [expected return given interest rate ] - [world interest rate]
        
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)				:: prior	! Prior belief
    REAL(DP), INTENT(IN)				:: rsrv_pre	! Incoming reserve
	REAL(DP), INTENT(IN)				:: capital	! Capital allocation
    REAL(DP), INTENT(IN)                :: yield    ! Normal times gross rate
        
    REAL(DP)                            :: participation
        
    !! LOCAL VARIABLES
    REAL(DP)                            :: phi_lbound
    REAL(DP)                            :: phi_ubound
    REAL(DP)                            :: res_ini
    REAL(DP)                            :: cdf_nocrisis
    REAL(DP)                            :: icdf_nocrisis
        
    !! PROGRAM        
    res_ini        = rsrv_ini(rsrv_pre, capital)
    
    IF ( res_ini .GE. ZERODP ) THEN
        
        phi_lbound      = phi_lb(rsrv_pre, capital, yield)
        phi_ubound      = phi_ub(rsrv_pre, capital, yield)
        
        IF ( phi_lbound .LT. phi_ubound ) THEN
            cdf_nocrisis    = cdf_h(phi_ubound,prior)  - cdf_h(phi_lbound,prior)
            icdf_nocrisis   = icdf_h(phi_ubound,prior) - icdf_h(phi_lbound,prior)
            
            participation   = icdf_nocrisis                                     &
                                + yield*(cdf_nocrisis - icdf_nocrisis)          &
                                + yield_crisis(rsrv_pre, capital, yield)        &
                                        *(ONEDP - cdf_nocrisis)                 &
                                - R
        ELSE
            participation   = - R - ONEDP   ! any value less than - R
        END IF
    
    ELSE
        participation   = - R - ONEDP       ! any value less than - R
    END IF
    
    
END FUNCTION participation

ELEMENTAL FUNCTION rsrv_ini(rsrv_pre, capital)
	!! Returns the reserves allocation
    !! given capital and incoming reserves
    
    !! R_1 =  1 + R_0 - K
	
	! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    
    REAL(DP)                :: rsrv_ini
    
	! PROGRAM
    rsrv_ini            = rsrv_pre + ONEDP - capital
    
END FUNCTION rsrv_ini

ELEMENTAL FUNCTION yield_crisis(rsrv_pre, capital, yield)
	!! Returns the crisis interest rate = (1 + r_L)
    !! assuming a feasible contract
    
    !! 1 + r_L = min (1+r_N, theta*(R_0 + lambda*K))
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: yield_crisis !  ( 1+ r_L)
    
	! PROGRAM
    yield_crisis            = MIN(yield, THETA*(rsrv_ini(rsrv_pre, capital) &
                                                + LAMBDA*capital))
    
END FUNCTION yield_crisis

ELEMENTAL FUNCTION liq_crisis(rsrv_pre, capital, yield)
	!! Returns the crisis liquidation level \ell_{SS}
    !! assuming a feasible contract
    
    !! \ell_{SS} = max {(1+r_L) - R_1, 0}/(\lambda K)
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: liq_crisis
    
    
	! PROGRAM
    liq_crisis              = ( yield_crisis(rsrv_pre, capital, yield)          & 
                                - rsrv_ini(rsrv_pre, capital)                   &
                               )/(LAMBDA*capital)
    
    liq_crisis              = MAX (ZERODP, liq_crisis)
END FUNCTION liq_crisis

ELEMENTAL FUNCTION phi_liq(rsrv_pre, capital, yield)
	!! Returns the lower bound on RC2 solvency states
    !! assuming a feasible contract
    
    !! Below phi_liq, RC2 is violated: too much owed.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_liq
    
    
	! PROGRAM
    phi_liq                 =  -( (A-ONEDP)*capital + rsrv_pre - (yield - ONEDP) )      &
                                / (yield - ONEDP)
    phi_liq                 = MIN (ONEDP, MAX (ZERODP, phi_liq))
    
END FUNCTION phi_liq

ELEMENTAL FUNCTION liq_max(rsrv_pre, capital, yield, shock)
	!! Returns the upper bound feasible time-1 liquidation
    !! given an interim shock
    !! assuming a feasible contract
    
    !! Liquidating above liq_max, RC2 is violated: too much owed.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate  ( 1+ r_N)
    REAL(DP), INTENT(IN)    :: shock        ! Liquidity shock (phi)
    
    REAL(DP)                :: liq_max
        
	! PROGRAM
    liq_max                 =  ( (A-ONEDP)*capital + rsrv_pre - (yield - ONEDP) )           &
                                /( (A-LAMBDA)*capital )                                     &
                                + ( (yield - ONEDP)*shock )                                 &
                                /( (A-LAMBDA)*capital )                                     
    
    liq_max                 = MIN (ONEDP, liq_max)
    
END FUNCTION liq_max

ELEMENTAL FUNCTION phi_up(rsrv_pre, capital, yield)
	!! Returns the "upper" bound on RC1 solvency states
    !! assuming a feasible contract
    
    !! Above phi_up, RC1 is violated: too much owed.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_up
    
    ! LOCAL VARIABLES
    INTEGER                 :: i
    REAL(DP), DIMENSION (N_PHI) :: liq_max_vec  ! R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI
    
    
	! PROGRAM
    FORALL ( i=1:N_PHI ) 
            ! compute shortfall in the interim assuming maximum liquidation
            ! R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI
            liq_max_vec(i)  = rsrv_ini(rsrv_pre, capital)                               &
                              + LAMBDA*liq_max(rsrv_pre, capital, yield, PHI_GRID(i))   &
                                      *capital                                          &
                              - PHI_GRID(i)
    END FORALL
    
    IF ( ALL(liq_max_vec .GE. ZERODP) ) THEN
        phi_up              = ONEDP
    ELSE IF ( ALL(liq_max_vec .LT. ZERODP) ) THEN
        phi_up              = ZERODP
    ELSE
        ! choose the largest phi with solvency
        ! i.e. the phi with the lowest positive interim leftover
        ! ( R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI >= 0 0 )
        phi_up              = MAXVAL(PHI_GRID, liq_max_vec .GE. ZERODP)
    END IF
    
END FUNCTION phi_up

ELEMENTAL FUNCTION phi_lo(rsrv_pre, capital, yield)
	!! Returns the "lower" bound on RC1 solvency states
    !! assuming a feasible contract
    
    !! Below phi_lo, RC1 is violated.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_lo
    
    ! LOCAL VARIABLES
    INTEGER                     :: i
    REAL(DP), DIMENSION (N_PHI) :: liq_max_vec  ! R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI
    
    
	! PROGRAM
    FORALL ( i=1:N_PHI ) 
            ! compute shortfall in the interim assuming maximum liquidation
            ! R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI
            liq_max_vec(i)  = rsrv_ini(rsrv_pre, capital)                               &
                              + LAMBDA*liq_max(rsrv_pre, capital, yield, PHI_GRID(i))   &
                                      *capital                                          &
                              - PHI_GRID(i)
    END FORALL
    
    IF ( ALL(liq_max_vec .GE. ZERODP) ) THEN
        phi_lo              = ZERODP
    ELSE IF ( ALL(liq_max_vec .LT. ZERODP) ) THEN
        phi_lo              = ONEDP
    ELSE
        ! choose the largest phi with solvency
        ! i.e. the phi with the largest positive interim leftover
        ! arg_max (R_1 + LAMBDA*LIQ_MAX(PHI)*K - PHI >= 0 )
        phi_lo              = MINVAL(PHI_GRID, liq_max_vec .GE. ZERODP)
    END IF
    
END FUNCTION phi_lo

ELEMENTAL FUNCTION phi_min(phi_liq, phi_lo)
	!! Returns the "lower" bound on solvency states
    !! assuming a feasible contract
    
    !! Below phi_min = max(phi_liq,phi_lo), RC1 OR RC2 is violated.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: phi_liq      ! lower bound for RC2 solvency
    REAL(DP), INTENT(IN)    :: phi_lo       ! lower bound for RC1 solvency
    
    REAL(DP)                :: phi_min
     
    
	! PROGRAM 
    phi_min                 = MAX(phi_liq,phi_lo)
    
END FUNCTION phi_min

ELEMENTAL FUNCTION phi_max(phi_up)
	!! Returns the upper bound on solvency states
    !! assuming a feasible contract
    
    !! Above phi_max = min(phi_up, 1.0), RC1 OR RC2 is violated.
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: phi_up       ! upper bound for RC1 solvency
    
    REAL(DP)                :: phi_max
     
    
	! PROGRAM   
    phi_max                 = MIN(phi_up,ONEDP)
    
END FUNCTION phi_max

ELEMENTAL FUNCTION phi_nodefault(rsrv_pre, capital, yield)
	!! Returns the lower bound on non-default states
    !! assuming a feasible contract
    
    !! Below phi_nodefault = phi_N, crisis occurs strategically = 'default'
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_nodefault
    
   
    
	! PROGRAM 
    phi_nodefault           = ( ( yield - yield_crisis(rsrv_pre, capital, yield) )      &
                                - (A-LAMBDA)*capital                                    &
                                            *liq_crisis(rsrv_pre, capital, yield)       &
                              )/(yield - ONEDP)

    phi_nodefault           = MIN(rsrv_ini(rsrv_pre, capital),phi_nodefault)    
    !phi_nodefault           = MIN(rsrv_ini(rsrv_pre, capital),MAX(phi_nodefault,ZERODP)) 
    
END FUNCTION phi_nodefault

ELEMENTAL FUNCTION phi_default(rsrv_pre, capital, yield)
	!! Returns the upper bound on non-default states
    !! assuming a feasible contract
    
    !! Above phi_default = phi_S, crisis occurs strategically = 'default'
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_default
    
   
    
	! PROGRAM 
    phi_default             = ( ( (A - LAMBDA)/LAMBDA )*rsrv_ini(rsrv_pre, capital)     &
                                + (A-LAMBDA)*capital                                    &
                                            *liq_crisis(rsrv_pre, capital, yield)       &
                                - ( yield - yield_crisis(rsrv_pre, capital, yield) )    &
                              )/( ((A - LAMBDA)/LAMBDA) - (yield - ONEDP) )

    phi_default             = MAX (rsrv_ini(rsrv_pre, capital), phi_default)    
    
END FUNCTION phi_default

ELEMENTAL FUNCTION phi_lb(rsrv_pre, capital, yield)
	!! Returns the lower bound on NORMAL TIMES states
    !! assuming a feasible contract
    
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_lb
    
   
    
	! PROGRAM 
    phi_lb                  = MAX (                                         &
                                phi_min(phi_liq(rsrv_pre, capital, yield),  &
                                        phi_lo(rsrv_pre, capital, yield)),  &
                                phi_nodefault(rsrv_pre, capital, yield)     &
                                    )
   
    
END FUNCTION phi_lb

ELEMENTAL FUNCTION phi_ub(rsrv_pre, capital, yield)
	!! Returns the upper bound on NORMAL TIMES states
    !! assuming a feasible contract
    
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    
    REAL(DP)                :: phi_ub
    
   
    
	! PROGRAM 
    phi_ub                  = MIN (                                         &
                                phi_max(phi_up(rsrv_pre, capital, yield)),  &
                                phi_default(rsrv_pre, capital, yield)       &
                                    )
   
    
END FUNCTION phi_ub

ELEMENTAL FUNCTION net_output_crisis(rsrv_pre, capital, yield)
	!! Returns the CRISIS NET OUTPUT = MAX RESERVES SAVINGS POSSIBLE
    !! assuming a feasible contract
    
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate (1+ r_N)
    
    REAL(DP)                :: net_output_crisis
    
    ! LOCAL VARIABLES
	REAL(DP)                :: res_ini
    REAL(DP)                :: ret_crisis
    REAL(DP)                :: net_ouput
    
	! PROGRAM
    res_ini                 = rsrv_ini(rsrv_pre, capital)
    ret_crisis              = yield_crisis(rsrv_pre, capital, yield)
    
    net_output_crisis       = A*capital + res_ini - ret_crisis                      &
                                - (A-LAMBDA)*capital                                &
                                            *liq_crisis(rsrv_pre, capital, yield)
    
END FUNCTION net_output_crisis

ELEMENTAL FUNCTION cons_crisis(rsrv_pre, capital, yield, rsrv_fin)
	!! Returns the CRISIS TIMES consumption
    !! assuming a feasible contract
    
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    REAL(DP), INTENT(IN)    :: rsrv_fin     ! final reserve allocation (savings)
    
    REAL(DP)                :: cons_crisis
    
   
	! PROGRAM
    cons_crisis             = net_output_crisis(rsrv_pre, capital, yield)            &
                                        - rsrv_fin
    
END FUNCTION cons_crisis

ELEMENTAL FUNCTION net_output(rsrv_pre, capital, yield, shock)
	!! Returns the  NET OUTPUT = MAX RESERVES SAVINGS POSSIBLE
    !! assuming a feasible contract
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    REAL(DP), INTENT(IN)    :: shock        ! liquidity shock (phi)
    
    REAL(DP)                :: net_output
    
    ! LOCAL VARIABLES
	REAL(DP)                :: res_ini
    REAL(DP)                :: liq_hat
    
	! PROGRAM
    res_ini                 = rsrv_ini(rsrv_pre, capital)
    
    net_output              =  A*capital + res_ini - ONEDP              &
                                - (yield - ONEDP)*(1 - shock)           &
                                - ((A-LAMBDA)/LAMBDA)                   &
                                    *MAX(shock - res_ini , ZERODP)      
    
                                
    
END FUNCTION net_output

ELEMENTAL FUNCTION cons_fin(rsrv_pre, capital, yield, shock, rsrv_fin)
	!! Returns the FINAL CONSUMPTION IN NORMAL TIMES
    !! assuming a feasible contract
	
    ! INPUT & OUTPUT
    REAL(DP), INTENT(IN)    :: rsrv_pre     ! initial reserve allocation
    REAL(DP), INTENT(IN)    :: capital      ! capital allocation
    REAL(DP), INTENT(IN)    :: yield        ! normal times interest rate ( 1+ r_N)
    REAL(DP), INTENT(IN)    :: shock        ! liquidity shock (phi)
    REAL(DP), INTENT(IN)    :: rsrv_fin     ! final reserve allocation (savings)
    
    REAL(DP)                :: cons_fin
    
  
    cons_fin                = net_output(rsrv_pre, capital, yield, shock) &
                                - rsrv_fin
    
END FUNCTION cons_fin
END MODULE toolkit